# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import functools
import inspect
import math
import operator
from builtins import bool, float, int, tuple  # noqa: A004 shadowing a Python built-in
from typing import (
    Any,
    Callable,
    Final,
    Generic,
    ParamSpec,
    Tuple,
    TypeAlias,
    TypeVar,
    Union,
    cast,
    overload,
)

import numpy as np
from numpy import float32, float64, int8, int16, int32, int64, uint8, uint16, uint32, uint64

from gt4py._core import definitions as core_defs
from gt4py.next import common, named_collections
from gt4py.next.common import Dimension, Field  # noqa: F401 [unused-import] for TYPE_BUILTINS
from gt4py.next.iterator import runtime
from gt4py.next.type_system import type_specifications as ts


__all__ = [  # noqa: RUF022 [unsorted-dunder-all] # type: ignore[attr-defined]
    "bool",
    "float",
    "float32",
    "float64",
    "int",
    "int8",
    "int16",
    "int32",
    "int64",
    "tuple",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "IndexType",
    "neighbor_sum",
    "max_over",
    "min_over",
    "where",
    "broadcast",
    "astype",
    "abs",
    "neg",
    "sin",
    "cos",
    "tan",
    "arcsin",
    "arccos",
    "arctan",
    "sinh",
    "cosh",
    "tanh",
    "arcsinh",
    "arccosh",
    "arctanh",
    "sqrt",
    "exp",
    "log",
    "gamma",
    "cbrt",
    "floor",
    "ceil",
    "trunc",
    "minimum",
    "maximum",
    "fmod",
    "power",
    "isfinite",
    "isinf",
    "isnan",
]


PYTHON_TYPE_BUILTINS = [bool, int, float, tuple]
PYTHON_TYPE_BUILTIN_NAMES = [t.__name__ for t in PYTHON_TYPE_BUILTINS]

TYPE_BUILTINS = [
    common.Field,
    common.Dimension,
    int8,
    uint8,
    int16,
    uint16,
    int32,
    uint32,
    int64,
    uint64,
    float32,
    float64,
    *PYTHON_TYPE_BUILTINS,
]  # TODO(tehrengruber): validate matches iterator.builtins.TYPE_BUILTINS?

TYPE_BUILTIN_NAMES = [t.__name__ for t in TYPE_BUILTINS]

# Be aware: Type aliases are not fully supported in the frontend yet, e.g. `IndexType(1)` will not
# work.
IndexType: TypeAlias = int32

TYPE_ALIAS_NAMES = ["IndexType"]

_P = ParamSpec("_P")
_R = TypeVar("_R")


def _type_conversion_helper(t: type) -> type[ts.TypeSpec] | tuple[type[ts.TypeSpec], ...]:
    if t is common.Field:
        return ts.FieldType
    elif t is common.Dimension:
        return ts.DimensionType
    elif t is FieldOffset:
        return ts.OffsetType
    elif t is common.Connectivity:
        return ts.OffsetType
    elif t is core_defs.ScalarT:
        return ts.ScalarType
    elif t is common.Domain:
        return ts.DomainType
    elif t is type:
        return (
            ts.ConstructorType
        )  # our type of type is currently represented by the type constructor function
    elif t is Tuple or (hasattr(t, "__origin__") and t.__origin__ is tuple):
        return ts.TupleType
    elif hasattr(t, "__origin__") and t.__origin__ is Union:
        types = [_type_conversion_helper(e) for e in t.__args__]  # type: ignore[attr-defined]
        assert all(type(t) is type and issubclass(t, ts.TypeSpec) for t in types)
        return cast(tuple[type[ts.TypeSpec], ...], tuple(types))  # `cast` to break the recursion
    elif t in named_collections.CUSTOM_NAMED_COLLECTION_TYPES:
        return ts.NamedCollectionType
    else:
        raise AssertionError("Illegal type encountered.")


def _type_conversion(t: type) -> ts.DeferredType:
    return ts.DeferredType(constraint=_type_conversion_helper(t))


@dataclasses.dataclass(frozen=True)
class BuiltInFunction(Generic[_R, _P]):
    name: str = dataclasses.field(init=False)
    # `function` can be used to provide a default implementation for all `Field` implementations,
    # e.g. a fused multiply add could have a default implementation as a*b+c, but an optimized implementation for a specific `Field`
    function: Callable[_P, _R]

    def __post_init__(self) -> None:
        object.__setattr__(self, "name", f"{self.function.__module__}.{self.function.__name__}")
        object.__setattr__(self, "__doc__", self.function.__doc__)

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _R:
        impl = self.dispatch(*args)
        return impl(*args, **kwargs)

    def dispatch(self, *args: Any) -> Callable[_P, _R]:
        arg_types = tuple(type(arg) for arg in args)
        for atype in arg_types:
            # current strategy is to select the implementation of the first arg that supports the operation
            # TODO: define a strategy that converts or prevents conversion
            if (dispatcher := getattr(atype, "__gt_builtin_func__", None)) is not None and (
                op_func := dispatcher(self)
            ) is not NotImplemented:
                return op_func
        else:
            return self.function

    def __gt_type__(self) -> ts.FunctionType:
        signature = inspect.signature(self.function)
        params = signature.parameters

        return ts.FunctionType(
            pos_only_args=[
                _type_conversion(param.annotation)
                for param in params.values()
                if param.kind == inspect.Parameter.POSITIONAL_ONLY
            ],
            pos_or_kw_args={
                k: _type_conversion(v.annotation)
                for k, v in params.items()
                if v.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
            },
            kw_only_args={
                k: _type_conversion(v.annotation)
                for k, v in params.items()
                if v.kind == inspect.Parameter.KEYWORD_ONLY
            },
            returns=_type_conversion(signature.return_annotation),
        )


CondT = TypeVar("CondT", bound=Union[common.Field, common.Domain])
FieldT1 = TypeVar(
    "FieldT1",
    bound=Union[common.Field, core_defs.Scalar, Tuple, named_collections.CustomNamedCollection],
)
FieldT2 = TypeVar(
    "FieldT2",
    bound=Union[common.Field, core_defs.Scalar, Tuple, named_collections.CustomNamedCollection],
)


class WhereBuiltinFunction(
    BuiltInFunction[_R, [CondT, FieldT1, FieldT2]],
    Generic[_R, CondT, FieldT1, FieldT2],
):
    @overload  # type: ignore[override] # this technically clashes with the superclass
    def __call__(  # type: ignore[overload-overlap] # without both overloads mypy raises false positives
        self,
        cond: CondT,
        true_field: common.Field | core_defs.ScalarT,
        false_field: common.Field | core_defs.ScalarT,
    ) -> common.Field: ...

    @overload
    def __call__(
        self,
        cond: CondT,
        true_field: Tuple | core_defs.ScalarT,
        false_field: Tuple | core_defs.ScalarT,
    ) -> Tuple: ...

    def __call__(self, cond: CondT, true_field: FieldT1, false_field: FieldT2) -> _R:  # type: ignore[misc] # supposedly this signature does not accept all the possible args allowed by the overloads ??
        if isinstance(true_field, tuple) or isinstance(false_field, tuple):
            if not (isinstance(true_field, tuple) and isinstance(false_field, tuple)):
                raise ValueError(
                    # TODO(havogt) find a strategy to unify parsing and embedded error messages
                    f"Either both or none can be tuple in '{true_field=}' and '{false_field=}'."
                )
            if len(true_field) != len(false_field):
                raise ValueError(
                    "Tuple of different size not allowed."
                )  # TODO(havogt) find a strategy to unify parsing and embedded error messages
            return tuple(self(cond, t, f) for t, f in zip(true_field, false_field))  # type: ignore[return-value] # `tuple` is not `_R`
        return super().__call__(cond, true_field, false_field)


@BuiltInFunction
def neighbor_sum(field: common.Field, /, axis: common.Dimension) -> common.Field:
    raise NotImplementedError()


@BuiltInFunction
def max_over(field: common.Field, /, axis: common.Dimension) -> common.Field:
    raise NotImplementedError()


@BuiltInFunction
def min_over(field: common.Field, /, axis: common.Dimension) -> common.Field:
    raise NotImplementedError()


@BuiltInFunction
def broadcast(
    field: common.Field | core_defs.ScalarT, dims: tuple[common.Dimension, ...], /
) -> common.Field:
    assert core_defs.is_scalar_type(
        field
    )  # default implementation for scalars, Fields are handled via dispatch
    # TODO(havogt) implement with FunctionField, the workaround is to ignore broadcasting on scalars as they broadcast automatically, but we lose the check for compatible dimensions
    return field  # type: ignore[return-value] # see comment above


@WhereBuiltinFunction
def where(
    mask: common.Field,
    true_field: common.Field | core_defs.ScalarT | Tuple | named_collections.CustomNamedCollection,
    false_field: common.Field | core_defs.ScalarT | Tuple | named_collections.CustomNamedCollection,
    /,
) -> common.Field | Tuple:
    raise NotImplementedError()


@overload
def astype(value: common.Field, type_: type, /) -> common.Field: ...


@overload
def astype(value: core_defs.ScalarT, type_: type, /) -> core_defs.ScalarT: ...


@overload
def astype(value: Tuple, type_: type, /) -> Tuple: ...


@BuiltInFunction
def astype(
    value: common.Field | core_defs.ScalarT | Tuple, type_: type, /
) -> common.Field | core_defs.ScalarT | Tuple:
    if isinstance(value, tuple):
        return tuple(astype(v, type_) for v in value)
    # default implementation for scalars, Fields are handled via dispatch
    assert core_defs.is_scalar_type(value)
    return core_defs.dtype(type_).scalar_type(value)


_UNARY_MATH_NUMBER_BUILTIN_IMPL: Final = {"abs": abs, "neg": operator.neg}
UNARY_MATH_NUMBER_BUILTIN_NAMES: Final = [*_UNARY_MATH_NUMBER_BUILTIN_IMPL.keys()]

_UNARY_MATH_FP_BUILTIN_IMPL: Final = {
    "sin": math.sin,
    "cos": math.cos,
    "tan": math.tan,
    "arcsin": math.asin,
    "arccos": math.acos,
    "arctan": math.atan,
    "sinh": math.sinh,
    "cosh": math.cosh,
    "tanh": math.tanh,
    "arcsinh": math.asinh,
    "arccosh": math.acosh,
    "arctanh": math.atanh,
    "sqrt": math.sqrt,
    "exp": math.exp,
    "log": math.log,
    "gamma": math.gamma,
    "cbrt": math.cbrt if hasattr(math, "cbrt") else np.cbrt,  # match.cbrt() only added in 3.11
    "floor": math.floor,
    "ceil": math.ceil,
    "trunc": math.trunc,
}
UNARY_MATH_FP_BUILTIN_NAMES: Final = [*_UNARY_MATH_FP_BUILTIN_IMPL.keys()]

_UNARY_MATH_FP_PREDICATE_BUILTIN_IMPL: Final = {
    "isfinite": math.isfinite,
    "isinf": math.isinf,
    "isnan": math.isnan,
}
UNARY_MATH_FP_PREDICATE_BUILTIN_NAMES: Final = [*_UNARY_MATH_FP_PREDICATE_BUILTIN_IMPL.keys()]


def _make_unary_math_builtin(name: str) -> BuiltInFunction:
    _math_builtin = (
        _UNARY_MATH_NUMBER_BUILTIN_IMPL
        | _UNARY_MATH_FP_BUILTIN_IMPL
        | _UNARY_MATH_FP_PREDICATE_BUILTIN_IMPL
    )[name]

    def impl(value: common.Field | core_defs.ScalarT, /) -> common.Field | core_defs.ScalarT:
        # TODO(havogt): enable tests in `test_math_builtin_execution.py`
        assert core_defs.is_scalar_type(
            value
        )  # default implementation for scalars, Fields are handled via dispatch

        return cast(common.Field | core_defs.ScalarT, _math_builtin(value))  # type: ignore[operator, arg-type] # calling a function of unknown type; trunc not supported for all types

    impl.__name__ = name
    return BuiltInFunction(impl)


abs = _make_unary_math_builtin("abs")  # noqa: A001 [shadowing]
neg = _make_unary_math_builtin("neg")
sin = _make_unary_math_builtin("sin")
cos = _make_unary_math_builtin("cos")
tan = _make_unary_math_builtin("tan")
arcsin = _make_unary_math_builtin("arcsin")
arccos = _make_unary_math_builtin("arccos")
arctan = _make_unary_math_builtin("arctan")
sinh = _make_unary_math_builtin("sinh")
cosh = _make_unary_math_builtin("cosh")
tanh = _make_unary_math_builtin("tanh")
arcsinh = _make_unary_math_builtin("arcsinh")
arccosh = _make_unary_math_builtin("arccosh")
arctanh = _make_unary_math_builtin("arctanh")
sqrt = _make_unary_math_builtin("sqrt")
exp = _make_unary_math_builtin("exp")
log = _make_unary_math_builtin("log")
gamma = _make_unary_math_builtin("gamma")
cbrt = _make_unary_math_builtin("cbrt")
floor = _make_unary_math_builtin("floor")
ceil = _make_unary_math_builtin("ceil")
trunc = _make_unary_math_builtin("trunc")
isfinite = _make_unary_math_builtin("isfinite")
isinf = _make_unary_math_builtin("isinf")
isnan = _make_unary_math_builtin("isnan")


BINARY_MATH_NUMBER_BUILTIN_NAMES = ["minimum", "maximum", "fmod", "power"]


def _make_binary_math_builtin(name: str) -> BuiltInFunction:
    def impl(
        lhs: common.Field | core_defs.ScalarT, rhs: common.Field | core_defs.ScalarT, /
    ) -> common.Field | core_defs.ScalarT:
        # default implementation for scalars, Fields are handled via dispatch
        assert core_defs.is_scalar_type(lhs)
        assert core_defs.is_scalar_type(rhs)
        return getattr(np, name)(lhs, rhs)

    impl.__name__ = name
    return BuiltInFunction(impl)


minimum = _make_binary_math_builtin("minimum")
maximum = _make_binary_math_builtin("maximum")
fmod = _make_binary_math_builtin("fmod")
power = _make_binary_math_builtin("power")


MATH_BUILTIN_NAMES = (
    UNARY_MATH_NUMBER_BUILTIN_NAMES
    + UNARY_MATH_FP_BUILTIN_NAMES
    + UNARY_MATH_FP_PREDICATE_BUILTIN_NAMES
    + BINARY_MATH_NUMBER_BUILTIN_NAMES
)

FUN_BUILTIN_NAMES = [
    "neighbor_sum",
    "max_over",
    "min_over",
    "broadcast",
    "where",
    "astype",
    *MATH_BUILTIN_NAMES,
]

assert all(isinstance(globals()[f], BuiltInFunction) for f in FUN_BUILTIN_NAMES), (
    "Missing builtin function"
)

BUILTIN_NAMES = TYPE_BUILTIN_NAMES + FUN_BUILTIN_NAMES

BUILTINS = {name: globals()[name] for name in BUILTIN_NAMES}

should_export = (set(BUILTIN_NAMES) | set(TYPE_ALIAS_NAMES)) - {"Dimension", "Field"}
actual_export = set(__all__)
assert (diff := should_export - actual_export) == set(), (
    f"Missing symbol(s) in 'fbuiltins.__all__': {diff}"
)
assert (diff := actual_export - should_export) == set(), (
    f"Symbol(s) exported but not defined in 'fbuiltins': {diff}"
)


# TODO(tehrengruber): FieldOffset and runtime.Offset are not an exact conceptual
#  match. Revisit if we want to continue subclassing here. If we split
#  them also check whether Dimension should continue to be the shared or define
#  guidelines for decision.
@dataclasses.dataclass(frozen=True)
class FieldOffset(runtime.Offset):
    source: common.Dimension
    target: tuple[common.Dimension] | tuple[common.Dimension, common.Dimension]

    @functools.cached_property
    def _cache(self) -> dict:
        return {}

    def __post_init__(self) -> None:
        if len(self.target) == 2 and self.target[1].kind != common.DimensionKind.LOCAL:
            raise ValueError("Second dimension in offset must be a local dimension.")

    def __gt_type__(self) -> ts.OffsetType:
        return ts.OffsetType(source=self.source, target=self.target)

    def __getitem__(self, offset: int) -> common.Connectivity:
        """Serve as a connectivity factory."""
        from gt4py.next import embedded  # avoid circular import

        assert isinstance(self.value, str)
        current_offset_provider = embedded.context.get_offset_provider(None)
        assert current_offset_provider is not None
        offset_definition = common.get_offset(current_offset_provider, self.value)

        connectivity: common.Connectivity
        if isinstance(offset_definition, common.Dimension):
            connectivity = common.CartesianConnectivity(offset_definition, offset)
        elif isinstance(offset_definition, common.Connectivity):
            assert common.is_neighbor_connectivity(offset_definition)
            named_index = common.NamedIndex(self.target[-1], offset)
            connectivity = offset_definition[named_index]
        else:
            raise NotImplementedError()

        return connectivity

    def as_connectivity_field(self) -> common.Connectivity:
        """Convert to connectivity field using the offset providers in current embedded execution context."""
        from gt4py.next import embedded  # avoid circular import

        assert isinstance(self.value, str)
        current_offset_provider = embedded.context.get_offset_provider(None)
        assert current_offset_provider is not None
        offset_definition = common.get_offset(current_offset_provider, self.value)

        cache_key = id(offset_definition)
        if (connectivity := self._cache.get(cache_key, None)) is None:
            if isinstance(offset_definition, common.Connectivity):
                connectivity = offset_definition
            else:
                raise NotImplementedError()

            self._cache[cache_key] = connectivity

        return connectivity



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import ast
import copy


class UnchainComparesPass(ast.NodeTransformer):
    """
    Turn comparison chain into a tree of comparisons linked by ``&`` operations.

    ``a > b < c == d ...`` is turned into ``((a > b) & ((b < c) & (c == d))`` and so on.
    The resulting AST will have only ``Comparison`` nodes with to operands and one operation.
    That is to say with the same structure as FOAST ``Comparison`` nodes.

    Examples:
    ---------
    >>> import inspect
    >>> def example(a, b, c, d):
    ...     return a > b < c == d
    >>> print(ast.unparse(UnchainComparesPass.apply(ast.parse(inspect.getsource(example)))))
    def example(a, b, c, d):
        return (a > b) & ((b < c) & (c == d))
    """

    @classmethod
    def apply(cls, node: ast.AST) -> ast.AST:
        return cls().visit(node)

    def visit_Compare(self, node: ast.Compare) -> ast.Compare | ast.BinOp:
        # stopping case: single comparison a <op> b
        if len(node.comparators) == 1:
            return node

        # left leaf of the new tree: ``a < b``
        # example: ``a < b > c > d``
        left_leaf = ast.Compare(
            ops=node.ops[0:1], left=node.left, comparators=node.comparators[0:1]
        )
        ast.copy_location(left_leaf, node)

        # the remainder of the chain -> right branch of the new tree
        # example: ``b > c > d``
        remaining_chain = copy.deepcopy(node)
        remaining_chain.left = remaining_chain.comparators.pop(0)
        remaining_chain.ops.pop(0)

        # create the tree root
        # example: ``(a < b) & ((b > c) & (c > d))``
        root = ast.BinOp(
            op=ast.BitAnd(),
            left=left_leaf,
            right=self.visit(remaining_chain),  # example: recursively visit ``b > c > d``
        )
        ast.copy_location(root, node)

        return root



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import enum
import os
import subprocess

# We rely on setuptools internal distutils - which is on
# a long term support track per the setuptools team
from setuptools._distutils.ccompiler import new_compiler
from setuptools._distutils.sysconfig import customize_compiler

from gt4py._core import definitions as core_defs


class CxxCompilerName(enum.Enum):
    DEFAULT = "unknown"
    GNU = "gcc"
    CLANG = "clang"
    APPLE_CLANG = "apple-clang"
    INTEL = "icx"


@dataclasses.dataclass(frozen=True)
class CxxCompilerDefaults:
    name: CxxCompilerName
    """Name identifier of the compiler"""
    open_mp_flag: str
    """OpenMP flag expected on a default install of the compiler"""
    enable_openmp: bool
    """Allow OpenMP acceleration"""
    cxx_compile_flags: str
    """Cxx compile flags"""


def cxx_compiler_defaults(optimization_level: str) -> CxxCompilerDefaults:
    """Return a set of defaults for the compiler flags"""

    # Get the compiler - relying on setuptools
    ccompiler = new_compiler()
    customize_compiler(ccompiler)

    # Defaults
    name = CxxCompilerName.DEFAULT
    open_mp_flags = "-fopenmp"
    cxx_flags = ""
    enable_openmp = True

    # FMA is deactivated by default when running -O0
    if optimization_level == "0":
        cxx_flags += "-ffp-contract=off"

    # Query the compiler version string
    try:
        compiler_exe_fullpath = ccompiler.compiler_cxx[0]
        r = subprocess.run([compiler_exe_fullpath, "--version"], capture_output=True, text=True)
        version_name_on_cli = r.stdout.split("\n")[0]
    except AttributeError:
        version_name_on_cli = "default"
    if "gcc" in version_name_on_cli.lower():
        name = CxxCompilerName.GNU
    elif "icx" in version_name_on_cli.lower() or "icpx" in version_name_on_cli.lower():
        name = CxxCompilerName.INTEL
        open_mp_flags = "-qopenmp"
    elif "apple clang" in version_name_on_cli.lower():
        # By default Apple Clang doesn't have OpenMP installed,
        # so we remove the flag and disallow OpenMP altogether
        name = CxxCompilerName.APPLE_CLANG
        open_mp_flags = ""
        enable_openmp = False
    elif "clang" in version_name_on_cli.lower():
        name = CxxCompilerName.CLANG

    return CxxCompilerDefaults(
        name,
        open_mp_flags,
        enable_openmp,
        cxx_flags.strip(),  # be overly cautious for old GCC bad behavior
    )


class GPUCompilerName(enum.Enum):
    NONE = "none"
    NVCC = "nvcc"
    ROCM = "rocm"


@dataclasses.dataclass(frozen=True)
class GPUConfiguration:
    name: GPUCompilerName
    """Name identifier of the compiler"""
    gpu_compile_flags: list[str]
    """Compile flags for device code"""
    binary_path: str
    """Path to binaries for GPU compiler & tools"""
    include_path: str
    """Path to includes for GPU runtime"""
    library_path: str
    """Path to libraries for GPU runtime"""
    arch: str | None
    """Device architecture (None will auto-detect)"""
    host_compiler: str | None
    """Host compiler (None will use CXX)"""


def gpu_configuration(optimization_level: str) -> GPUConfiguration:
    """Retrieve all information to compile and execute GPU device code"""

    # We base our GPU configuration on the CUDA_HOME environment variable
    # as we have mostly supported CUDA historically
    cuda_root = os.environ.get(
        "CUDA_HOME", os.environ.get("CUDA_PATH", os.path.abspath("/usr/local/cuda"))
    )

    # Based on `cupy` we detect a ROCM capable compiler
    if core_defs.CUPY_DEVICE_TYPE == core_defs.DeviceType.ROCM:
        name = GPUCompilerName.ROCM
        library_path = os.path.join(cuda_root, "lib")
    else:
        name = GPUCompilerName.NVCC
        library_path = os.path.join(cuda_root, "lib64")

    # Default arguments for GPU source code
    gpu_compile_flags_default = ""
    if optimization_level == "0":
        # When running -O0 we deactivate FMA
        gpu_compile_flags_default = "-fmad=false"

    extra_cuda_compile_args = os.environ.get(
        "GT4PY_CARTESIAN_EXTRA_CUDA_COMPILE_ARGS", gpu_compile_flags_default
    )
    gpu_compile_flags = extra_cuda_compile_args.split(" ") if extra_cuda_compile_args else []

    return GPUConfiguration(
        name=name,
        gpu_compile_flags=gpu_compile_flags,
        binary_path=os.path.join(cuda_root, "bin"),
        include_path=os.path.join(cuda_root, "include"),
        library_path=library_path,
        arch=os.environ.get("CUDA_ARCH", None),
        host_compiler=os.environ.get("CUDA_HOST_CXX", None),
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import functools
from typing import Callable, Literal, Optional, cast

from gt4py.next import common, utils as next_utils
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    domain_utils,
    ir_makers as im,
    misc as ir_utils_misc,
)
from gt4py.next.iterator.ir_utils.domain_utils import SymbolicDomain
from gt4py.next.iterator.transforms import cse, infer_domain, inline_lambdas
from gt4py.next.iterator.type_system import inference as type_inference
from gt4py.next.type_system import type_info, type_specifications as ts


def select_elems_by_domain(
    select_domain: SymbolicDomain,
    target: itir.Expr,
    source: itir.Expr,
    domains: tuple[SymbolicDomain, ...],
):
    """
    Select all elements of possibly nested tuples for the given domain.

    Returns (non-nested) tuples of the selected elements and the corresponding targets.
    """
    new_targets = []
    new_els = []
    for i, el_domain in enumerate(domains):
        current_target = im.tuple_get(i, target)
        current_source = im.tuple_get(i, source)
        if isinstance(el_domain, tuple):
            more_targets, more_els = select_elems_by_domain(
                select_domain, current_target, current_source, el_domain
            )
            new_els.extend(more_els)
            new_targets.extend(more_targets)
        else:
            assert isinstance(el_domain, SymbolicDomain)
            if el_domain == select_domain:
                new_targets.append(current_target)
                new_els.append(current_source)
    return new_targets, new_els


def _set_at_for_domain(stmt: itir.SetAt, domain: SymbolicDomain) -> itir.SetAt:
    """Extract all elements with given domain into a new `SetAt` statement."""
    tuple_expr = stmt.expr
    targets, expr_els = select_elems_by_domain(
        domain, stmt.target, tuple_expr, stmt.expr.annex.domain
    )
    new_expr = im.make_tuple(*expr_els)
    new_expr.annex.domain = domain

    return itir.SetAt(expr=new_expr, domain=domain.as_expr(), target=im.make_tuple(*targets))


def _populate_and_homogenize_domains(stmts: list[itir.Stmt]) -> list[itir.Stmt]:
    """
    Splits `SetAt` statements with multiple domains into multiple statements.

    `SetAt`s have a single domain therefore the target domains have to be the same.
    We support nested tuples by flattening them and recombining them into non-nested tuples.
    """
    new_stmts: list[itir.Stmt] = []
    for stmt in stmts:
        if isinstance(stmt, itir.SetAt):
            assert hasattr(stmt.expr.annex, "domain")
            # ordered set for reproducibility
            domains: dict[
                SymbolicDomain | Literal[infer_domain.DomainAccessDescriptor.NEVER], None
            ] = dict.fromkeys(next_utils.flatten_nested_tuple(stmt.expr.annex.domain))
            # don't count NEVER as a different domain
            domains.pop(infer_domain.DomainAccessDescriptor.NEVER, None)
            distinct_domains: list[domain_utils.SymbolicDomain] = list(
                cast(dict[SymbolicDomain, None], domains)
            )
            if len(distinct_domains) == 1:
                new_stmts.append(
                    itir.SetAt(
                        expr=stmt.expr,
                        domain=distinct_domains[0].as_expr(),  # insert concrete domain
                        target=stmt.target,
                    )
                )
            else:
                new_stmts.extend(_set_at_for_domain(stmt, domain) for domain in distinct_domains)
        elif isinstance(stmt, itir.IfStmt):
            new_stmts.append(
                itir.IfStmt(
                    cond=stmt.cond,
                    true_branch=_populate_and_homogenize_domains(stmt.true_branch),
                    false_branch=_populate_and_homogenize_domains(stmt.false_branch),
                )
            )
        else:
            raise ValueError("Expected `itir.SetAt` or `itir.IfStmt`.")

    return new_stmts


def _is_as_fieldop_of_scan(expr: itir.Expr) -> bool:
    return (
        cpm.is_applied_as_fieldop(expr)
        and isinstance(expr.fun, itir.FunCall)
        and cpm.is_call_to(expr.fun.args[0], "scan")
    )


def _transform_if(
    stmt: itir.Stmt,
    declarations: list[itir.Temporary],
    uids: next_utils.IDGeneratorPool,
) -> Optional[list[itir.Stmt]]:
    if isinstance(stmt, itir.SetAt) and cpm.is_call_to(stmt.expr, "if_"):
        cond, true_val, false_val = stmt.expr.args
        return [
            itir.IfStmt(
                cond=cond,
                true_branch=_transform_stmt(
                    itir.SetAt(target=stmt.target, expr=true_val, domain=stmt.domain),
                    declarations,
                    uids,
                ),
                false_branch=_transform_stmt(
                    itir.SetAt(target=stmt.target, expr=false_val, domain=stmt.domain),
                    declarations,
                    uids,
                ),
            )
        ]
    return None


def _transform_by_pattern(
    stmt: itir.Stmt,
    predicate: Callable[[itir.Expr, int], bool],
    declarations: list[itir.Temporary],
    uids: next_utils.IDGeneratorPool,
) -> Optional[list[itir.Stmt]]:
    if not isinstance(stmt, itir.SetAt):
        return None

    # hide projector from extraction
    projector, expr = ir_utils_misc.extract_projector(stmt.expr)

    new_expr, extracted_fields, _ = cse.extract_subexpression(
        expr,
        predicate=predicate,
        prefixed_uids=uids["__tmp_subexpr"],
        # TODO(tehrengruber): extracting the deepest expression first would allow us to fuse
        #  the extracted expressions resulting in fewer kernel calls & better data-locality.
        #  Extracting multiple expressions deepest-first is however not supported right now.
        # deepest_expr_first=True  # noqa: ERA001
    )

    if extracted_fields:
        tmp_stmts: list[itir.Stmt] = []

        # for each extracted expression generate:
        #  - one or more `Temporary` declarations (depending on whether the expression is a field
        #    or a tuple thereof)
        #  - one `SetAt` statement that materializes the expression into the temporary
        for tmp_sym, tmp_expr in extracted_fields.items():
            assert isinstance(tmp_expr.type, ts.TypeSpec)
            tmp_names: str | tuple[str | tuple, ...] = type_info.apply_to_primitive_constituents(
                lambda x: next(uids["__tmp"]),
                tmp_expr.type,
                tuple_constructor=lambda *elements: tuple(elements),
            )
            tmp_dtypes: (
                ts.ScalarType | ts.ListType | tuple[ts.ScalarType | ts.ListType | tuple, ...]
            ) = type_info.apply_to_primitive_constituents(
                type_info.extract_dtype,
                tmp_expr.type,
                tuple_constructor=lambda *elements: elements,
            )

            tmp_domains: SymbolicDomain | tuple[SymbolicDomain | tuple, ...] = tmp_expr.annex.domain

            if cpm.is_applied_as_fieldop(tmp_expr):
                # In this case all tuple elements have the same size (or will be `NEVER`).
                # Create the tuple structure with that domain.
                domain = list(
                    set(next_utils.flatten_nested_tuple((tmp_domains,)))
                    - {infer_domain.DomainAccessDescriptor.NEVER}  # type: ignore[arg-type] # type should always be `SymbolicDomain`
                )
                assert len(domain) == 1
                # this is the domain used as initial value in the tuple construction below
                tmp_domains = domain[0]

            def get_domain(
                _, path: tuple[int, ...]
            ) -> domain_utils.SymbolicDomain:  # function only used inside loop
                domain = functools.reduce(
                    lambda var, idx: var[idx] if isinstance(var, tuple) else var,
                    path,
                    tmp_domains,  # noqa: B023
                )
                assert isinstance(domain, domain_utils.SymbolicDomain)
                return domain

            # The following propagates the domains to the tuple structure of `tmp_expr.type`.
            # `tmp_domains` might not have this structure because domain inference was not able to infer the tuple structure.
            tmp_domains = type_info.apply_to_primitive_constituents(
                get_domain,
                tmp_expr.type,
                with_path_arg=True,
                tuple_constructor=lambda *elements: tuple(elements),
            )

            declarations.extend(
                itir.Temporary(id=tmp_name, domain=domain.as_expr(), dtype=dtype)
                for tmp_name, domain, dtype in zip(
                    next_utils.flatten_nested_tuple((tmp_names,)),
                    next_utils.flatten_nested_tuple((tmp_domains,)),
                    next_utils.flatten_nested_tuple((tmp_dtypes,)),
                    strict=True,
                )
            )

            # if the expr is a field this just gives a simple `itir.SymRef`, otherwise we generate a
            #  `make_tuple` expression.
            target_expr: itir.Expr = next_utils.tree_map(
                lambda name, domain: im.ref(name, annex={"domain": domain}),
                result_collection_constructor=lambda _, elts: im.make_tuple(*elts),
            )(tmp_names, tmp_domains)  # type: ignore[assignment]  # typing of tree_map does not reflect action of `result_collection_constructor` yet

            # note: the let would be removed automatically by the `cse.extract_subexpression`, but
            # we remove it here for readability & debuggability.
            new_expr = inline_lambdas.inline_lambda(
                im.let(tmp_sym, target_expr)(new_expr), opcount_preserving=False
            )

            # TODO(tehrengruber): _transform_stmt not needed if deepest_expr_first=True
            tmp_stmts.extend(
                _transform_stmt(
                    itir.SetAt(
                        target=target_expr,
                        # The domain is populated later in `_populate_and_homogenize_domains`
                        # at this point the `SetAt` contains possibly expressions on different domains.
                        domain=im.ref("UNPOPULATED"),
                        expr=tmp_expr,
                    ),
                    declarations,
                    uids,
                )
            )

        if projector is not None:
            # add the projector back
            domain = new_expr.annex.domain
            new_expr = im.call(projector)(new_expr)
            new_expr.annex.domain = domain

        return [*tmp_stmts, itir.SetAt(target=stmt.target, domain=stmt.domain, expr=new_expr)]
    return None


def _transform_stmt(
    stmt: itir.Stmt,
    declarations: list[itir.Temporary],
    uids: next_utils.IDGeneratorPool,
) -> list[itir.Stmt]:
    unprocessed_stmts: list[itir.Stmt] = [stmt]
    stmts: list[itir.Stmt] = []

    transforms: list[Callable] = [
        # transform `if_` call into `IfStmt`
        _transform_if,
        # extract applied `as_fieldop` to top-level
        functools.partial(
            _transform_by_pattern, predicate=lambda expr, _: cpm.is_applied_as_fieldop(expr)
        ),
        # extract if_ call to the top-level
        functools.partial(
            _transform_by_pattern, predicate=lambda expr, _: cpm.is_call_to(expr, "if_")
        ),
    ]

    while unprocessed_stmts:
        stmt = unprocessed_stmts.pop(0)

        did_transform = False
        for transform in transforms:
            transformed_stmts = transform(stmt=stmt, declarations=declarations, uids=uids)
            if transformed_stmts:
                unprocessed_stmts = [*transformed_stmts, *unprocessed_stmts]
                did_transform = True
                break

        # no transformation occurred
        if not did_transform:
            stmts.append(stmt)

    stmts = _populate_and_homogenize_domains(stmts)

    return stmts


def create_global_tmps(
    program: itir.Program,
    offset_provider: common.OffsetProvider | common.OffsetProviderType,
    #: A dictionary mapping axes names to their length. See :func:`infer_domain.infer_expr` for
    #: more details.
    symbolic_domain_sizes: Optional[dict[str, itir.Expr]] = None,
    *,
    uids: next_utils.IDGeneratorPool,
) -> itir.Program:
    """
    Given an `itir.Program` create temporaries for intermediate values.

    This pass looks at all `as_fieldop` calls and transforms field-typed subexpressions of its
    arguments into temporaries.
    """
    program = infer_domain.infer_program(
        program,
        offset_provider=offset_provider,
        symbolic_domain_sizes=symbolic_domain_sizes,
        # Previous passes are allowed to create expressions without domains, but we must not
        # overwrite other domains here. Instead, only reinfer expressions without a domain. We must
        # not overwrite them since e.g. a `concat_where` expression might be rewritten into an
        # `if_(cond, tb, fb)` expression where re-inference might extend the domain of tb and fb.
        # See :class:`infer_domain.infer_expr` for details.
        keep_existing_domains=True,
    )
    program = type_inference.infer(
        program, offset_provider_type=common.offset_provider_to_type(offset_provider)
    )

    declarations = program.declarations.copy()
    new_body = []

    for stmt in program.body:
        assert isinstance(stmt, itir.SetAt)
        new_body.extend(_transform_stmt(stmt, uids=uids, declarations=declarations))

    return itir.Program(
        id=program.id,
        function_definitions=program.function_definitions,
        params=program.params,
        declarations=declarations,
        body=new_body,
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from gt4py.next.ffront.decorator import (
    embedded_program_call_context as embedded_program_call_context,
    program_call_context as program_call_context,
)
from gt4py.next.otf.compiled_program import (
    compile_variant_hook as compile_variant_hook,
    compiled_program_call_context as compiled_program_call_context,
)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""
GT4Py-NEXT - Performance portable and composable weather & climate stencils.

This module deviates from the project coding style (Google Python style) in the following way:

[style guide link](https://google.github.io/styleguide/pyguide.html#22-imports):

Fnctions and classes are imported from other modules in order to explicitly re-export them
to create a streamlined user experience. `from module import *` can be used but only if the
module in question is a submodule, defines `__all__` and exports many public API objects.
"""

# ruff: noqa: F401
from .._core.definitions import CUPY_DEVICE_TYPE, Device, DeviceType, is_scalar_type
from . import common, ffront, iterator, program_processors, typing
from .common import (
    Connectivity,
    Dimension,
    DimensionKind,
    Dims,
    Domain,
    Field,
    GridType,
    UnitRange,
    domain,
    unit_range,
)
from .constructors import FieldConstructor, as_connectivity, as_field, empty, full, ones, zeros
from .embedded import (  # Just for registering field implementations
    nd_array_field as _nd_array_field,
)
from .ffront import fbuiltins
from .ffront.decorator import field_operator, program, scan_operator
from .ffront.fbuiltins import (
    FieldOffset,
    IndexType,
    abs,  # noqa: A004 # shadowing
    arccos,
    arccosh,
    arcsin,
    arcsinh,
    arctan,
    arctanh,
    astype,
    bool,  # noqa: A004 # shadowing
    broadcast,
    cbrt,
    ceil,
    cos,
    cosh,
    exp,
    float,  # noqa: A004 # shadowing
    float32,
    float64,
    floor,
    fmod,
    gamma,
    int,  # noqa: A004 # shadowing
    int8,
    int16,
    int32,
    int64,
    isfinite,
    isinf,
    isnan,
    log,
    max_over,
    maximum,
    min_over,
    minimum,
    neg,
    neighbor_sum,
    power,
    sin,
    sinh,
    sqrt,
    tan,
    tanh,
    trunc,
    tuple,  # noqa: A004 # shadowing
    uint8,
    uint16,
    uint32,
    uint64,
    where,
)
from .otf.compiled_program import wait_for_compilation
from .program_processors.runners.gtfn import (
    run_gtfn_cached as gtfn_cpu,
    run_gtfn_gpu_cached as gtfn_gpu,
)
from .program_processors.runners.roundtrip import default as itir_python


__all__ = [  # noqa: RUF022 [unsorted-dunder-all]
    # submodules
    "common",
    "ffront",
    "iterator",
    "program_processors",
    "typing",
    # from _core.definitions
    "CUPY_DEVICE_TYPE",
    "Device",
    "DeviceType",
    "is_scalar_type",
    # from common
    "Dimension",
    "DimensionKind",
    "Dims",
    "Field",
    "Connectivity",
    "GridType",
    "domain",
    "Domain",
    "unit_range",
    "UnitRange",
    # from constructors
    "FieldConstructor",
    "empty",
    "zeros",
    "ones",
    "full",
    "as_field",
    "as_connectivity",
    # from ffront
    "FieldOffset",
    "field_operator",
    "program",
    "scan_operator",
    # from otf
    "wait_for_compilation",
    # from program_processor
    "gtfn_cpu",
    "gtfn_gpu",
    "itir_python",
    # from fbuiltins
    "FieldOffset",
    "IndexType",
    "abs",
    "arccos",
    "arccosh",
    "arcsin",
    "arcsinh",
    "arctan",
    "arctanh",
    "astype",
    "bool",
    "broadcast",
    "cbrt",
    "ceil",
    "cos",
    "cosh",
    "exp",
    "float",
    "float32",
    "float64",
    "floor",
    "fmod",
    "gamma",
    "int",
    "int8",
    "int16",
    "int32",
    "int64",
    "isfinite",
    "isinf",
    "isnan",
    "log",
    "max_over",
    "min_over",
    "maximum",
    "minimum",
    "neg",
    "neighbor_sum",
    "power",
    "sin",
    "sinh",
    "sqrt",
    "tan",
    "tanh",
    "trunc",
    "tuple",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "where",
]

assert not (diff := set(fbuiltins.__all__) - set(__all__)), (
    f"public fbuiltins member(s) not exported: {diff}"
)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Hashable
from typing import Any, Callable, Protocol, TypeGuard, cast

import array_api_compat
import numpy as np

from gt4py._core import definitions as core_defs


try:
    import cupy
except ImportError:
    cupy = None


class _ArrayDType(Protocol):
    # The only requirement for array API dtypes is equality comparability.
    def __eq__(self, other: Any) -> bool: ...


class ArrayNamespace(Hashable, Protocol):
    """
    Currently only a subset of the Array API standard namespace with functions relevant for array creation.

    TODO(havogt): replace by ... or make it more complete and put next to NDArrayObject definition.
    """

    def empty(
        self, shape: tuple[int, ...], *, dtype: Any | None = None, device: Any | None = None
    ) -> core_defs.NDArrayObject: ...

    def zeros(
        self, shape: tuple[int, ...], *, dtype: Any | None = None, device: Any | None = None
    ) -> core_defs.NDArrayObject: ...

    def ones(
        self, shape: tuple[int, ...], *, dtype: Any | None = None, device: Any | None = None
    ) -> core_defs.NDArrayObject: ...

    def full(
        self,
        shape: tuple[int, ...],
        fill_value: Any,
        *,
        dtype: Any | None = None,
        device: Any | None = None,
    ) -> core_defs.NDArrayObject: ...

    def asarray(
        self,
        obj: Any,
        /,
        *,
        dtype: Any | None = None,
        device: Any | None = None,
        copy: bool | None = None,
    ) -> core_defs.NDArrayObject: ...

    bool: _ArrayDType
    int8: _ArrayDType
    int16: _ArrayDType
    int32: _ArrayDType
    int64: _ArrayDType
    uint8: _ArrayDType
    uint16: _ArrayDType
    uint32: _ArrayDType
    uint64: _ArrayDType
    float32: _ArrayDType
    float64: _ArrayDType


def is_array_namespace(obj: Any) -> TypeGuard[ArrayNamespace]:
    """
    Check whether `obj` (structurally) is a namespace of the array API.

    See description in 'ArrayNamespace'.
    """

    return (
        hasattr(obj, "empty")
        and hasattr(obj, "zeros")
        and hasattr(obj, "ones")
        and hasattr(obj, "full")
        and hasattr(obj, "asarray")
        and (hasattr(obj, "bool") or hasattr(obj, "bool_"))  # for NumPy 1.x compatibility
        and hasattr(obj, "int8")
        and hasattr(obj, "int16")
        and hasattr(obj, "int32")
        and hasattr(obj, "int64")
        and hasattr(obj, "uint8")
        and hasattr(obj, "uint16")
        and hasattr(obj, "uint32")
        and hasattr(obj, "uint64")
        and hasattr(obj, "float32")
        and hasattr(obj, "float64")
    )


def array_namespace(array: core_defs.NDArrayObject) -> ArrayNamespace:
    return cast(ArrayNamespace, array_api_compat.array_namespace(array))


def _numpy_device_translator(device: core_defs.Device | None) -> Any:
    if device is None:
        return None
    if device.device_type == core_defs.DeviceType.CPU:
        return None  # or Literal['cpu']
    raise ValueError(f"NumPy does not support device type {device.device_type}.")


# Currently private as we only support a concrete set of array namespaces.
_device_translation_registry: dict[ArrayNamespace, Callable[[core_defs.Device | None], Any]] = {
    cast(ArrayNamespace, np): _numpy_device_translator
}
"""
Registry for mappings from array namespaces to device_translators.

Device translators are functions mapping GT4Py 'Device' objects to the corresponding device objects for the given array namespace.
It's responsibility of the translator to check if 'Device' is valid for the given array namespace and raise a 'ValueError' if not.
"""

if cupy is not None:

    def _cupy_device_translator(device: core_defs.Device | None) -> Any:
        if device is None:
            return None
        if device.device_type != core_defs.CUPY_DEVICE_TYPE:
            raise ValueError(
                f"CuPy only supports GPU devices, got device type {device.device_type}."
            )
        return cupy.cuda.Device(device.device_id)

    _device_translation_registry[cupy] = _cupy_device_translator


def get_device_translator(array_ns: ArrayNamespace) -> Callable[[core_defs.Device | None], Any]:
    """
    Returns a mapping from a GT4Py 'Device' to the corresponding device object for the given array namespace.
    """
    try:
        return _device_translation_registry[array_ns]
    except KeyError:
        raise ValueError(
            f"No device translator registered for array namespace {array_ns}."
        ) from None



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import textwrap
from typing import Any, Optional, Sequence, TypeAlias, TypeVar, cast

import gt4py.next.ffront.field_operator_ast as foast
from gt4py import eve
from gt4py.eve import NodeTranslator, NodeVisitor, traits
from gt4py.next import errors
from gt4py.next.common import Dimension, DimensionKind, promote_dims
from gt4py.next.ffront import (
    dialect_ast_enums,
    experimental,
    fbuiltins,
    type_info as ti_ffront,
    type_specifications as ts_ffront,
)
from gt4py.next.ffront.foast_passes import utils as foast_utils
from gt4py.next.iterator import builtins
from gt4py.next.type_system import type_info, type_specifications as ts, type_translation


OperatorNodeT = TypeVar("OperatorNodeT", bound=foast.LocatedNode)


def with_altered_scalar_kind(
    type_spec: ts.TypeSpec, new_scalar_kind: ts.ScalarKind
) -> ts.ScalarType | ts.FieldType:
    """
    Given a scalar or field type, return a type with different scalar kind.

    Examples:
    ---------
    >>> scalar_t = ts.ScalarType(kind=ts.ScalarKind.FLOAT64)
    >>> print(with_altered_scalar_kind(scalar_t, ts.ScalarKind.BOOL))
    bool

    >>> field_t = ts.FieldType(
    ...     dims=[Dimension(value="I")], dtype=ts.ScalarType(kind=ts.ScalarKind.FLOAT64)
    ... )
    >>> print(with_altered_scalar_kind(field_t, ts.ScalarKind.FLOAT32))
    Field[[I], float32]
    """
    if isinstance(type_spec, ts.FieldType):
        return ts.FieldType(
            dims=type_spec.dims,
            dtype=with_altered_scalar_kind(type_spec.dtype, new_scalar_kind),
        )
    elif isinstance(type_spec, ts.ScalarType):
        return ts.ScalarType(kind=new_scalar_kind, shape=type_spec.shape)
    else:
        raise ValueError(f"Expected field or scalar type, got '{type_spec}'.")


def deduce_stmt_return_type(
    node: foast.BlockStmt, *, requires_unconditional_return: bool = True
) -> Optional[ts.TypeSpec]:
    """
    Deduce type of value returned inside a block statement.

    If `requires_unconditional_return` is true the function additionally ensures that the block
    statement unconditionally returns and raises an `AssertionError` if not.
    """
    conditional_return_type: Optional[ts.TypeSpec] = None

    for stmt in node.stmts:
        is_unconditional_return = False
        return_type: Optional[ts.TypeSpec]

        if isinstance(stmt, foast.Return):
            is_unconditional_return = True
            return_type = stmt.value.type
        elif isinstance(stmt, foast.IfStmt):
            return_types = (
                deduce_stmt_return_type(stmt.true_branch, requires_unconditional_return=False),
                deduce_stmt_return_type(stmt.false_branch, requires_unconditional_return=False),
            )
            # if both branches return
            if return_types[0] and return_types[1]:
                if return_types[0] == return_types[1]:
                    is_unconditional_return = True
                else:
                    raise errors.DSLError(
                        stmt.location,
                        "If statement contains return statements with inconsistent types:"
                        f"{return_types[0]} != {return_types[1]}",
                    )
            return_type = return_types[0] or return_types[1]
        elif isinstance(stmt, foast.BlockStmt):
            # just forward to nested BlockStmt
            return_type = deduce_stmt_return_type(
                stmt, requires_unconditional_return=requires_unconditional_return
            )
        elif isinstance(stmt, (foast.Assign, foast.TupleTargetAssign)):
            return_type = None
        else:
            raise AssertionError(f"Nodes of type '{type(stmt).__name__}' not supported.")

        if conditional_return_type and return_type and return_type != conditional_return_type:
            raise errors.DSLError(
                stmt.location,
                "If statement contains return statements with inconsistent types:"
                f"{conditional_return_type} != {conditional_return_type}",
            )

        if is_unconditional_return:  # found a statement that always returns
            assert return_type
            return return_type
        elif return_type:
            conditional_return_type = return_type

    if requires_unconditional_return:
        # If the node was constructed by the foast parsing we should never get here, but instead
        # we should have gotten an error there.
        raise AssertionError(
            "Malformed block statement: expected a return statement in this context, "
            "but none was found. Please submit a bug report."
        )

    return None


class FieldOperatorTypeDeductionCompletnessValidator(NodeVisitor):
    """Validate an FOAST expression is fully typed."""

    @classmethod
    def apply(cls, node: foast.LocatedNode) -> None:
        incomplete_nodes: list[foast.LocatedNode] = []
        cls().visit(node, incomplete_nodes=incomplete_nodes)

        if incomplete_nodes:
            raise AssertionError("'FOAST' expression is not fully typed.")

    def visit_LocatedNode(
        self, node: foast.LocatedNode, *, incomplete_nodes: list[foast.LocatedNode]
    ) -> None:
        self.generic_visit(node, incomplete_nodes=incomplete_nodes)

        if hasattr(node, "type") and not type_info.is_concrete(node.type):
            incomplete_nodes.append(node)


class FieldOperatorTypeDeduction(traits.VisitorWithSymbolTableTrait, NodeTranslator):
    """
    Deduce and check types of FOAST expressions and symbols.

    Examples:
    ---------
    >>> import ast
    >>> import typing
    >>> from gt4py.next import Field
    >>> from gt4py.next.ffront.source_utils import SourceDefinition, get_closure_vars_from_function
    >>> from gt4py.next.ffront.func_to_foast import FieldOperatorParser
    >>> IDim = Dimension("IDim")
    >>> def example(a: "Field[[IDim], float]", b: "Field[[IDim], float]"):
    ...     return a + b

    >>> source_definition = SourceDefinition.from_function(example)
    >>> closure_vars = get_closure_vars_from_function(example)
    >>> annotations = typing.get_type_hints(example)
    >>> untyped_fieldop = FieldOperatorParser(
    ...     source_definition=source_definition, closure_vars=closure_vars, annotations=annotations
    ... ).visit(ast.parse(source_definition.source).body[0])
    >>> untyped_fieldop.body.stmts[0].value.type
    DeferredType(constraint=None)

    >>> typed_fieldop = FieldOperatorTypeDeduction.apply(untyped_fieldop)
    >>> assert typed_fieldop.body.stmts[0].value.type == ts.FieldType(
    ...     dtype=ts.ScalarType(kind=ts.ScalarKind.FLOAT64), dims=[IDim]
    ... )
    """

    @classmethod
    def apply(cls, node: OperatorNodeT) -> OperatorNodeT:
        typed_foast_node = cls().visit(node)

        FieldOperatorTypeDeductionCompletnessValidator.apply(typed_foast_node)

        return typed_foast_node

    def visit_FunctionDefinition(
        self, node: foast.FunctionDefinition, **kwargs: Any
    ) -> foast.FunctionDefinition:
        new_params = self.visit(node.params, **kwargs)
        new_body = self.visit(node.body, **kwargs)
        new_closure_vars = self.visit(node.closure_vars, **kwargs)
        return_type = deduce_stmt_return_type(new_body)
        if not isinstance(return_type, (ts.DataType, ts.DeferredType, ts.VoidType)):
            raise errors.DSLError(
                node.location,
                f"Function must return 'DataType', 'DeferredType', or 'VoidType', got '{return_type}'.",
            )
        new_type = ts.FunctionType(
            pos_only_args=[],
            pos_or_kw_args={str(new_param.id): new_param.type for new_param in new_params},
            kw_only_args={},
            returns=return_type,
        )
        return foast.FunctionDefinition(
            id=node.id,
            params=new_params,
            body=new_body,
            closure_vars=new_closure_vars,
            type=new_type,
            location=node.location,
        )

    # TODO(tehrengruber): make sure all scalar arguments are lifted to 0-dim field args
    def visit_FieldOperator(self, node: foast.FieldOperator, **kwargs: Any) -> foast.FieldOperator:
        new_definition = self.visit(node.definition, **kwargs)
        return foast.FieldOperator(
            id=node.id,
            definition=new_definition,
            location=node.location,
            type=ts_ffront.FieldOperatorType(definition=new_definition.type),
        )

    def visit_ScanOperator(self, node: foast.ScanOperator, **kwargs: Any) -> foast.ScanOperator:
        new_axis = self.visit(node.axis, **kwargs)
        if not isinstance(new_axis.type, ts.DimensionType):
            raise errors.DSLError(
                node.location, f"Argument 'axis' to scan operator '{node.id}' must be a dimension."
            )
        if not new_axis.type.dim.kind == DimensionKind.VERTICAL:
            raise errors.DSLError(
                node.location,
                f"Argument 'axis' to scan operator '{node.id}' must be a vertical dimension.",
            )
        new_forward = self.visit(node.forward, **kwargs)
        if not new_forward.type.kind == ts.ScalarKind.BOOL:
            raise errors.DSLError(
                node.location, f"Argument 'forward' to scan operator '{node.id}' must be a boolean."
            )
        new_init = self.visit(node.init, **kwargs)
        if not all(
            type_info.is_arithmetic(type_) or type_info.is_logical(type_)
            for type_ in type_info.primitive_constituents(new_init.type)
        ):
            raise errors.DSLError(
                node.location,
                f"Argument 'init' to scan operator '{node.id}' must "
                "be an arithmetic type or a logical type or a composite of arithmetic and logical types.",
            )
        new_definition = self.visit(node.definition, **kwargs)
        new_def_type = new_definition.type
        carry_type = next(iter(new_def_type.pos_or_kw_args.values()))
        if new_init.type != new_def_type.returns:
            raise errors.DSLError(
                node.location,
                f"Argument 'init' to scan operator '{node.id}' must have same type as its return: "
                f"expected '{new_def_type.returns}', got '{new_init.type}'.",
            )
        elif new_init.type != carry_type:
            carry_arg_name = next(iter(new_def_type.pos_or_kw_args.keys()))
            raise errors.DSLError(
                node.location,
                f"Argument 'init' to scan operator '{node.id}' must have same type as '{carry_arg_name}' argument: "
                f"expected '{carry_type}', got '{new_init.type}'.",
            )

        new_type = ts_ffront.ScanOperatorType(axis=new_axis.type.dim, definition=new_def_type)
        return foast.ScanOperator(
            id=node.id,
            axis=new_axis,
            forward=new_forward,
            init=new_init,
            definition=new_definition,
            type=new_type,
            location=node.location,
        )

    def visit_Name(self, node: foast.Name, **kwargs: Any) -> foast.Name:
        symtable = kwargs["symtable"]
        if node.id not in symtable or symtable[node.id].type is None:
            raise errors.DSLError(node.location, f"Undeclared symbol '{node.id}'.")

        symbol = symtable[node.id]
        return foast.Name(id=node.id, type=symbol.type, location=node.location)

    def visit_Assign(self, node: foast.Assign, **kwargs: Any) -> foast.Assign:
        new_value = node.value
        if not type_info.is_concrete(node.value.type):
            new_value = self.visit(node.value, **kwargs)
        new_target = self.visit(node.target, refine_type=new_value.type, **kwargs)
        return foast.Assign(target=new_target, value=new_value, location=node.location)

    def visit_TupleTargetAssign(
        self, node: foast.TupleTargetAssign, **kwargs: Any
    ) -> foast.TupleTargetAssign:
        TargetType: TypeAlias = list[foast.Starred | foast.Symbol]
        values = self.visit(node.value, **kwargs)

        if isinstance(values.type, ts.COLLECTION_TYPE_SPECS):
            num_elts: int = len(values.type.types)
            targets: TargetType = node.targets
            indices: list[tuple[int, int] | int] = foast_utils.compute_assign_indices(
                targets, num_elts
            )

            if not any(isinstance(i, tuple) for i in indices) and len(targets) != num_elts:
                raise errors.DSLError(
                    node.location, f"Too many values to unpack (expected {len(targets)})."
                )

            new_targets: TargetType = []
            new_type: ts.DataType
            for i, index in enumerate(indices):
                old_target = targets[i]

                if isinstance(index, tuple):
                    lower, upper = index
                    new_type = ts.TupleType(types=[t for t in values.type.types[lower:upper]])
                    new_target = foast.Starred(
                        id=foast.DataSymbol(
                            id=self.visit(old_target.id).id,
                            location=old_target.location,
                            type=new_type,
                        ),
                        type=new_type,
                        location=old_target.location,
                    )
                else:
                    new_type = values.type.types[index]
                    assert isinstance(new_type, ts.DataType)
                    new_target = self.visit(
                        old_target, refine_type=new_type, location=old_target.location, **kwargs
                    )

                new_target = self.visit(
                    new_target, refine_type=new_type, location=old_target.location, **kwargs
                )
                new_targets.append(new_target)
        else:
            raise errors.DSLError(
                node.location, f"Assignment value must be of type tuple, got '{values.type}'."
            )

        return foast.TupleTargetAssign(targets=new_targets, value=values, location=node.location)

    def visit_IfStmt(self, node: foast.IfStmt, **kwargs: Any) -> foast.IfStmt:
        symtable = kwargs["symtable"]

        new_true_branch = self.visit(node.true_branch, **kwargs)
        new_false_branch = self.visit(node.false_branch, **kwargs)
        new_node = foast.IfStmt(
            condition=self.visit(node.condition, **kwargs),
            true_branch=new_true_branch,
            false_branch=new_false_branch,
            location=node.location,
        )

        if not isinstance(new_node.condition.type, ts.ScalarType):
            raise errors.DSLError(
                node.location,
                f"Condition for 'if' must be scalar, got '{new_node.condition.type}' instead.",
            )

        if new_node.condition.type.kind != ts.ScalarKind.BOOL:
            raise errors.DSLError(
                node.location,
                "Condition for 'if' must be of boolean type, "
                f"got '{new_node.condition.type}' instead.",
            )

        for sym in node.annex.propagated_symbols.keys():
            if (true_type := new_true_branch.annex.symtable[sym].type) != (
                false_type := new_false_branch.annex.symtable[sym].type
            ):
                raise errors.DSLError(
                    node.location,
                    f"Inconsistent types between two branches for variable '{sym}': "
                    f"got types '{true_type}' and '{false_type}.",
                )
            # TODO: properly patch symtable (new node?)
            symtable[sym].type = new_node.annex.propagated_symbols[sym].type = (
                new_true_branch.annex.symtable[sym].type
            )

        return new_node

    def visit_Symbol(
        self, node: foast.Symbol, refine_type: Optional[ts.FieldType] = None, **kwargs: Any
    ) -> foast.Symbol:
        symtable = kwargs["symtable"]
        if refine_type:
            if not type_info.is_concretizable(node.type, to_type=refine_type):
                raise errors.DSLError(
                    node.location,
                    (
                        "Type inconsistency: expression was deduced to be "
                        f"of type '{refine_type}', instead of the expected type '{node.type}'."
                    ),
                )
            new_node: foast.Symbol = foast.Symbol(
                id=node.id, type=refine_type, location=node.location
            )
            symtable[new_node.id] = new_node
            return new_node
        return node

    def visit_Attribute(self, node: foast.Attribute, **kwargs: Any) -> foast.Attribute:
        new_value = self.visit(node.value, **kwargs)
        return foast.Attribute(
            value=new_value,
            attr=node.attr,
            location=node.location,
            type=getattr(new_value.type, node.attr),
        )

    def visit_Subscript(self, node: foast.Subscript, **kwargs: Any) -> foast.Subscript:
        new_value = self.visit(node.value, **kwargs)
        new_index = self.visit(node.index, **kwargs)
        new_type: Optional[ts.TypeSpec] = None

        match new_value.type:
            case ts.TupleType(types=types):
                try:
                    index = foast_utils.expr_to_index(node.index)
                except ValueError as ex:
                    raise errors.DSLError(
                        node.location,
                        f"Tuples need to be indexed with literal integers, got '{node.index}'.",
                    ) from ex
                new_type = types[index]
            case ts.OffsetType(source=source, target=(target1, target2)):
                if not target2.kind == DimensionKind.LOCAL:
                    raise errors.DSLError(
                        new_value.location, "Second dimension in offset must be a local dimension."
                    )
                new_type = ts.OffsetType(source=source, target=(target1,))
            case ts.OffsetType(source=source, target=(target,)):
                # for cartesian axes (e.g. I, J) the index of the subscript only
                #  signifies the displacement in the respective dimension,
                #  but does not change the target type.
                if source != target:
                    raise errors.DSLError(
                        new_value.location,
                        "Source and target must be equal for offsets with a single target.",
                    )
                new_type = new_value.type
            case ts.FieldType(dims=dims, dtype=dtype):
                # e.g. `field[LocalDim(42)]`
                new_type = ts.FieldType(
                    dims=[d for d in dims if d != new_index.type.dim],
                    dtype=dtype,
                )
            case _:
                raise errors.DSLError(
                    new_value.location, "Could not deduce type of subscript expression."
                )

        return foast.Subscript(
            value=new_value,
            index=new_index,
            type=new_type,
            location=node.location,
        )

    def visit_BinOp(self, node: foast.BinOp, **kwargs: Any) -> foast.BinOp:
        new_left = self.visit(node.left, **kwargs)
        new_right = self.visit(node.right, **kwargs)
        new_type = self._deduce_binop_type(node, left=new_left, right=new_right)
        return foast.BinOp(
            op=node.op, left=new_left, right=new_right, location=node.location, type=new_type
        )

    def visit_TernaryExpr(self, node: foast.TernaryExpr, **kwargs: Any) -> foast.TernaryExpr:
        new_condition = self.visit(node.condition, **kwargs)
        new_true_expr = self.visit(node.true_expr, **kwargs)
        new_false_expr = self.visit(node.false_expr, **kwargs)
        new_type = self._deduce_ternaryexpr_type(
            node, condition=new_condition, true_expr=new_true_expr, false_expr=new_false_expr
        )
        return foast.TernaryExpr(
            condition=new_condition,
            true_expr=new_true_expr,
            false_expr=new_false_expr,
            location=node.location,
            type=new_type,
        )

    def _deduce_ternaryexpr_type(
        self,
        node: foast.TernaryExpr,
        *,
        condition: foast.Expr,
        true_expr: foast.Expr,
        false_expr: foast.Expr,
    ) -> Optional[ts.TypeSpec]:
        if condition.type != ts.ScalarType(kind=ts.ScalarKind.BOOL):
            raise errors.DSLError(
                condition.location,
                f"Condition is of type '{condition.type}', should be of type 'bool'.",
            )

        if true_expr.type != false_expr.type:
            raise errors.DSLError(
                node.location,
                f"Left and right types are not the same: '{true_expr.type}' and '{false_expr.type}'",
            )
        return true_expr.type

    def visit_Compare(self, node: foast.Compare, **kwargs: Any) -> foast.Compare:
        new_left = self.visit(node.left, **kwargs)
        new_right = self.visit(node.right, **kwargs)
        new_type = self._deduce_compare_type(node, left=new_left, right=new_right)
        return foast.Compare(
            op=node.op, left=new_left, right=new_right, location=node.location, type=new_type
        )

    def _deduce_arithmetic_compare_type(
        self, node: foast.Compare, *, left: foast.Expr, right: foast.Expr, **kwargs: Any
    ) -> Optional[ts.TypeSpec]:
        # e.g. `1 < 2`
        self._check_operand_dtypes_match(node, left=left, right=right)

        try:
            # transform operands to have bool dtype and use regular promotion
            #  mechanism to handle dimension promotion
            return type_info.promote(
                with_altered_scalar_kind(left.type, ts.ScalarKind.BOOL),
                with_altered_scalar_kind(right.type, ts.ScalarKind.BOOL),
            )
        except ValueError as ex:
            raise errors.DSLError(
                node.location,
                f"Could not promote '{left.type}' and '{right.type}' to common type"
                f" in call to '{node.op}'.",
            ) from ex

    def _deduce_dimension_compare_type(
        self, node: foast.Compare, *, left: foast.Expr, right: foast.Expr, **kwargs: Any
    ) -> Optional[ts.TypeSpec]:
        # e.g. `IDim > 1`
        index_type = ts.ScalarType(
            kind=getattr(ts.ScalarKind, builtins.INTEGER_INDEX_BUILTIN.upper())
        )

        def error_msg(left: ts.TypeSpec, right: ts.TypeSpec) -> str:
            return f"Dimension comparison needs to be between a 'Dimension' and index of type '{index_type}', got '{left}' and '{right}'."

        if isinstance(left.type, ts.DimensionType):
            if not right.type == index_type:
                raise errors.DSLError(
                    right.location,
                    error_msg(left.type, right.type),
                )
            return ts.DomainType(dims=[left.type.dim])
        elif isinstance(right.type, ts.DimensionType):
            if not left.type == index_type:
                raise errors.DSLError(
                    left.location,
                    error_msg(left.type, right.type),
                )
            return ts.DomainType(dims=[right.type.dim])
        else:
            raise AssertionError()

    def _deduce_compare_type(
        self, node: foast.Compare, *, left: foast.Expr, right: foast.Expr, **kwargs: Any
    ) -> Optional[ts.TypeSpec]:
        # e.g. `1 < 1`
        if all(type_info.is_arithmetic(arg) for arg in (left.type, right.type)):
            return self._deduce_arithmetic_compare_type(node, left=left, right=right)
        # e.g. `IDim > 1`
        if any(isinstance(arg, ts.DimensionType) for arg in (left.type, right.type)):
            return self._deduce_dimension_compare_type(node, left=left, right=right)

        raise errors.DSLError(
            left.location,
            "Comparison operators can only be used between arithmetic types "
            "(scalars, fields) or between a dimension and an index type "
            "({builtins.INTEGER_INDEX_BUILTIN}).",
        )

    def _deduce_binop_type(
        self, node: foast.BinOp, *, left: foast.Expr, right: foast.Expr, **kwargs: Any
    ) -> Optional[ts.TypeSpec]:
        # e.g. `IDim+1`
        if (
            isinstance(left.type, ts.DimensionType)
            and isinstance(right.type, ts.ScalarType)
            and type_info.is_integral(right.type)
        ):
            return ts.OffsetType(source=left.type.dim, target=(left.type.dim,))
        if isinstance(left.type, ts.OffsetType):
            raise errors.DSLError(
                node.location, f"Type '{left.type}' can not be used in operator '{node.op}'."
            )

        logical_ops = {
            dialect_ast_enums.BinaryOperator.BIT_AND,
            dialect_ast_enums.BinaryOperator.BIT_OR,
            dialect_ast_enums.BinaryOperator.BIT_XOR,
        }

        err_msg = f"Unsupported operand type(s) for {node.op}: '{left.type}' and '{right.type}'."

        if isinstance(left.type, (ts.ScalarType, ts.FieldType)) and isinstance(
            right.type, (ts.ScalarType, ts.FieldType)
        ):
            is_compatible = (
                type_info.is_logical if node.op in logical_ops else type_info.is_arithmetic
            )
            for arg in (left, right):
                if not is_compatible(arg.type):
                    raise errors.DSLError(arg.location, err_msg)

            if node.op == dialect_ast_enums.BinaryOperator.POW:
                return left.type

            if node.op == dialect_ast_enums.BinaryOperator.MOD and not type_info.is_integral(
                right.type
            ):
                raise errors.DSLError(
                    arg.location,
                    f"Type '{right.type}' can not be used in operator '{node.op}', it only accepts 'int'.",
                )

            try:
                return type_info.promote(left.type, right.type)
            except ValueError as ex:
                raise errors.DSLError(
                    node.location,
                    f"Could not promote '{left.type}' and '{right.type}' to common type"
                    f" in call to '{node.op}'.",
                ) from ex
        elif isinstance(left.type, ts.DomainType) and isinstance(right.type, ts.DomainType):
            if node.op not in logical_ops:
                raise errors.DSLError(
                    node.location,
                    f"{err_msg} Operator "
                    f"must be one of {', '.join((str(op) for op in logical_ops))}.",
                )
            return ts.DomainType(dims=promote_dims(left.type.dims, right.type.dims))
        else:
            raise errors.DSLError(node.location, err_msg)

    def _check_operand_dtypes_match(
        self, node: foast.BinOp | foast.Compare, left: foast.Expr, right: foast.Expr
    ) -> None:
        # check dtypes match
        if not type_info.extract_dtype(left.type) == type_info.extract_dtype(right.type):
            raise errors.DSLError(
                node.location,
                f"Incompatible datatypes in operator '{node.op}': '{left.type}' and '{right.type}'.",
            )

    def visit_UnaryOp(self, node: foast.UnaryOp, **kwargs: Any) -> foast.UnaryOp:
        new_operand = self.visit(node.operand, **kwargs)
        is_compatible = (
            type_info.is_logical
            if node.op
            in [dialect_ast_enums.UnaryOperator.NOT, dialect_ast_enums.UnaryOperator.INVERT]
            else type_info.is_arithmetic
        )
        if not is_compatible(new_operand.type):
            raise errors.DSLError(
                node.location,
                f"Incompatible type for unary operator '{node.op}': '{new_operand.type}'.",
            )
        return foast.UnaryOp(
            op=node.op, operand=new_operand, location=node.location, type=new_operand.type
        )

    def visit_TupleExpr(self, node: foast.TupleExpr, **kwargs: Any) -> foast.TupleExpr:
        new_elts = self.visit(node.elts, **kwargs)
        new_type = ts.TupleType(types=[element.type for element in new_elts])
        return foast.TupleExpr(elts=new_elts, type=new_type, location=node.location)

    def visit_Call(self, node: foast.Call, **kwargs: Any) -> foast.Call:
        new_func = self.visit(node.func, **kwargs)
        new_args = self.visit(node.args, **kwargs)
        new_kwargs = self.visit(node.kwargs, **kwargs)

        func_type = new_func.type
        arg_types = [arg.type for arg in new_args]
        kwarg_types = {name: arg.type for name, arg in new_kwargs.items()}

        if isinstance(
            new_func.type,
            (
                ts.FunctionType,
                ts_ffront.FieldOperatorType,
                ts_ffront.ScanOperatorType,
                ts.ConstructorType,
            ),
        ):
            # Since we use the `id` attribute in the latter part of the toolchain ensure we
            # have the proper format here.
            if not isinstance(
                new_func,
                (foast.FunctionDefinition, foast.FieldOperator, foast.ScanOperator, foast.Name),
            ):
                raise errors.DSLError(node.location, "Functions can only be called directly.")
        elif isinstance(new_func.type, ts.FieldType):
            pass
        elif isinstance(new_func.type, ts.DimensionType):
            assert new_func.type.dim.kind == DimensionKind.LOCAL
            return foast.Call(
                func=new_func,
                args=new_args,
                kwargs=new_kwargs,
                location=node.location,
                type=ts.IndexType(dim=new_func.type.dim),
            )
        else:
            raise errors.DSLError(
                node.location,
                f"Expression of type '{new_func.type}' is not callable, must be a 'Function', 'FieldOperator', 'ScanOperator' or 'Field'.",
            )

        # ensure signature is valid
        try:
            type_info.accepts_args(
                func_type, with_args=arg_types, with_kwargs=kwarg_types, raise_exception=True
            )
        except ValueError as err:
            raise errors.DSLError(
                node.location, f"Invalid argument types in call to '{new_func}'.\n{err}"
            ) from err

        return_type = type_info.return_type(func_type, with_args=arg_types, with_kwargs=kwarg_types)

        new_node = foast.Call(
            func=new_func,
            args=new_args,
            kwargs=new_kwargs,
            location=node.location,
            type=return_type,
        )

        if (
            isinstance(new_func.type, ts.FunctionType)
            and isinstance(new_func, foast.Name)
            and new_func.id in fbuiltins.MATH_BUILTIN_NAMES
        ):
            return self._visit_math_built_in(new_node, **kwargs)
        elif (
            isinstance(new_func.type, ts.FunctionType)
            and not type_info.is_concrete(return_type)
            and isinstance(new_func, foast.Name)
            and new_func.id
            in (fbuiltins.FUN_BUILTIN_NAMES + experimental.EXPERIMENTAL_FUN_BUILTIN_NAMES)
        ):
            visitor = getattr(self, f"_visit_{new_func.id}")
            return visitor(new_node, **kwargs)

        return new_node

    def _visit_math_built_in(self, node: foast.Call, **kwargs: Any) -> foast.Call:
        func_name = cast(foast.Name, node.func).id

        # validate arguments
        error_msg_preamble = f"Incompatible argument in call to '{func_name}'."
        error_msg_for_validator = {
            type_info.is_arithmetic: "an arithmetic",
            type_info.is_floating_point: "a floating point",
        }
        if func_name in fbuiltins.UNARY_MATH_NUMBER_BUILTIN_NAMES:
            arg_validator = type_info.is_arithmetic
        elif func_name in fbuiltins.UNARY_MATH_FP_BUILTIN_NAMES:
            arg_validator = type_info.is_floating_point
        elif func_name in fbuiltins.UNARY_MATH_FP_PREDICATE_BUILTIN_NAMES:
            arg_validator = type_info.is_floating_point
        elif func_name in fbuiltins.BINARY_MATH_NUMBER_BUILTIN_NAMES:
            arg_validator = type_info.is_arithmetic
        else:
            raise AssertionError(f"Unknown math builtin '{func_name}'.")

        error_msgs = []
        for i, arg in enumerate(node.args):
            if not arg_validator(arg.type):
                error_msgs.append(
                    f"Expected {i}-th argument to be {error_msg_for_validator[arg_validator]} type, got '{arg.type}'."
                )
        if error_msgs:
            raise errors.DSLError(
                node.location,
                "\n".join([error_msg_preamble] + [f"  - {error}" for error in error_msgs]),
            )

        if func_name == "power" and all(type_info.is_integral(arg.type) for arg in node.args):
            print(f"Warning: return type of '{func_name}' might be inconsistent (not implemented).")

        # deduce return type
        return_type: Optional[ts.FieldType | ts.ScalarType] = None
        if (
            func_name
            in fbuiltins.UNARY_MATH_NUMBER_BUILTIN_NAMES + fbuiltins.UNARY_MATH_FP_BUILTIN_NAMES
        ):
            return_type = cast(ts.FieldType | ts.ScalarType, node.args[0].type)
        elif func_name in fbuiltins.UNARY_MATH_FP_PREDICATE_BUILTIN_NAMES:
            return_type = with_altered_scalar_kind(
                cast(ts.FieldType | ts.ScalarType, node.args[0].type), ts.ScalarKind.BOOL
            )
        elif func_name in fbuiltins.BINARY_MATH_NUMBER_BUILTIN_NAMES:
            try:
                return_type = type_info.promote(
                    *((cast(ts.FieldType | ts.ScalarType, arg.type)) for arg in node.args)
                )
            except ValueError as ex:
                raise errors.DSLError(node.location, error_msg_preamble) from ex
        else:
            raise AssertionError(f"Unknown math builtin '{func_name}'.")

        return foast.Call(
            func=node.func,
            args=node.args,
            kwargs=node.kwargs,
            location=node.location,
            type=return_type,
        )

    def _visit_reduction(self, node: foast.Call, **kwargs: Any) -> foast.Call:
        field_type = cast(ts.FieldType, node.args[0].type)
        reduction_dim = cast(ts.DimensionType, node.kwargs["axis"].type).dim
        # TODO: This code does not handle ellipses for dimensions. Fix it.
        assert field_type.dims is not ...
        if reduction_dim not in field_type.dims:
            field_dims_str = ", ".join(str(dim) for dim in field_type.dims)
            raise errors.DSLError(
                node.location,
                f"Incompatible field argument in call to '{node.func!s}'. "
                f"Expected a field with dimension '{reduction_dim}', got "
                f"'{field_dims_str}'.",
            )
        return_type = ts.FieldType(
            dims=[dim for dim in field_type.dims if dim != reduction_dim], dtype=field_type.dtype
        )

        return foast.Call(
            func=node.func,
            args=node.args,
            kwargs=node.kwargs,
            location=node.location,
            type=return_type,
        )

    def _visit_neighbor_sum(self, node: foast.Call, **kwargs: Any) -> foast.Call:
        return self._visit_reduction(node, **kwargs)

    def _visit_max_over(self, node: foast.Call, **kwargs: Any) -> foast.Call:
        return self._visit_reduction(node, **kwargs)

    def _visit_min_over(self, node: foast.Call, **kwargs: Any) -> foast.Call:
        return self._visit_reduction(node, **kwargs)

    def _visit_astype(self, node: foast.Call, **kwargs: Any) -> foast.Call:
        value, new_type_constructor = node.args
        assert isinstance(
            value.type, (ts.FieldType, ts.ScalarType, ts.TupleType)
        )  # already checked using generic mechanism

        # Note: the type to convert to is uniquely identified by its GT4Py type (`ConstructorType`),
        # not by e.g. its name.
        if not (
            isinstance(new_type_constructor.type, ts.ConstructorType)
            and isinstance(new_type_constructor.type.definition.returns, ts.ScalarType)
        ):
            raise errors.DSLError(
                node.location,
                f"Invalid call to 'astype': second argument must be a scalar type, got '{new_type_constructor}'.",
            )

        new_type = new_type_constructor.type.definition.returns

        return_type = type_info.apply_to_primitive_constituents(
            lambda primitive_type: with_altered_scalar_kind(primitive_type, new_type.kind),
            value.type,
        )
        assert isinstance(return_type, (ts.TupleType, ts.ScalarType, ts.FieldType))

        return foast.Call(
            func=node.func,
            args=node.args,
            kwargs=node.kwargs,
            type=return_type,
            location=node.location,
        )

    def _visit_as_offset(self, node: foast.Call, **kwargs: Any) -> foast.Call:
        arg_0 = node.args[0].type
        arg_1 = node.args[1].type
        assert isinstance(arg_0, ts.OffsetType)
        assert isinstance(arg_1, ts.FieldType)
        if not type_info.is_integral(arg_1):
            raise errors.DSLError(
                node.location,
                f"Incompatible argument in call to '{node.func!s}': "
                f"expected integer for offset field dtype, got '{arg_1.dtype}'. "
                f"{node.location}",
            )

        if arg_0.source not in arg_1.dims:
            raise errors.DSLError(
                node.location,
                f"Incompatible argument in call to '{node.func!s}': "
                f"'{arg_0.source}' not in list of offset field dimensions '{arg_1.dims}'. "
                f"{node.location}",
            )

        return foast.Call(
            func=node.func, args=node.args, kwargs=node.kwargs, type=arg_0, location=node.location
        )

    def _deduce_where_return_type(
        self,
        func_name: str,
        cond_dims: Sequence[Dimension],
        true_branch: ts.FieldType | ts.TupleType | ts.NamedCollectionType,
        false_branch: ts.FieldType | ts.TupleType | ts.NamedCollectionType,
        location: eve.SourceLocation,
    ) -> ts.FieldType | ts.TupleType | ts.NamedCollectionType:
        assert all(
            isinstance(el, (ts.FieldType, ts.ScalarType))
            for arg in (true_branch, false_branch)
            for el in type_info.primitive_constituents(arg)
        )

        # replace all primitive constituents by the same type, `ts.DeferredType()` for convenience,
        # to capture the structure of the two branches
        extract_structure = ti_ffront.tree_map_type(lambda x: ts.DeferredType(constraint=None))
        tb_structure = extract_structure(true_branch)
        fb_structure = extract_structure(false_branch)

        if tb_structure != fb_structure:
            raise errors.DSLError(
                location,
                f"Second and third argument to '{func_name}' must have the same tuple/collection "
                f"structure.\n"
                + textwrap.indent(
                    f"true branch:  '{true_branch}'\nfalse branch:  '{false_branch}'", " "
                ),
            )

        @ti_ffront.tree_map_type
        def deduce_return_type(
            tb: ts.FieldType | ts.ScalarType, fb: ts.FieldType | ts.ScalarType
        ) -> ts.FieldType:
            if (t_dtype := type_info.extract_dtype(tb)) != (f_dtype := type_info.extract_dtype(fb)):
                raise errors.DSLError(
                    location,
                    f"Field arguments to '{func_name}' must be of same dtype, got '{t_dtype}' != "
                    f"'{f_dtype}'.",
                )
            return_dims = promote_dims(cond_dims, type_info.extract_dims(type_info.promote(tb, fb)))
            return_type = ts.FieldType(dims=return_dims, dtype=t_dtype)
            return return_type

        return deduce_return_type(true_branch, false_branch)  # type: ignore[return-value]

    def _visit_where(self, node: foast.Call, **kwargs: Any) -> foast.Call:
        mask_type, true_branch_type, false_branch_type = (arg.type for arg in node.args)
        assert isinstance(mask_type, ts.FieldType)
        if not type_info.is_logical(mask_type):
            raise errors.DSLError(
                node.location,
                f"Incompatible argument in call to '{node.func!s}': expected "
                f"a field with dtype 'bool', got '{mask_type}'.",
            )
        return_type = self._deduce_where_return_type(
            "where",
            mask_type.dims,
            true_branch_type,  # type: ignore[arg-type]
            false_branch_type,  # type: ignore[arg-type]
            node.location,
        )

        return foast.Call(
            func=node.func,
            args=node.args,
            kwargs=node.kwargs,
            type=return_type,
            location=node.location,
        )

    def _visit_concat_where(self, node: foast.Call, **kwargs: Any) -> foast.Call:
        cond_type, true_branch_type, false_branch_type = (arg.type for arg in node.args)

        assert isinstance(cond_type, ts.DomainType)
        return_type = self._deduce_where_return_type(
            "concat_where",
            cond_type.dims,
            true_branch_type,  # type: ignore[arg-type]
            false_branch_type,  # type: ignore[arg-type]
            node.location,
        )

        return foast.Call(
            func=node.func,
            args=node.args,
            kwargs=node.kwargs,
            type=return_type,
            location=node.location,
        )

    def _visit_broadcast(self, node: foast.Call, **kwargs: Any) -> foast.Call:
        arg_type = cast(ts.FieldType | ts.ScalarType, node.args[0].type)
        broadcast_dims_expr = cast(foast.TupleExpr, node.args[1]).elts

        if any([not (isinstance(elt.type, ts.DimensionType)) for elt in broadcast_dims_expr]):
            raise errors.DSLError(
                node.location,
                f"Incompatible broadcast dimension type in '{node.func!s}': expected "
                f"all broadcast dimensions to be of type 'Dimension'.",
            )

        broadcast_dims = [cast(ts.DimensionType, elt.type).dim for elt in broadcast_dims_expr]

        if not set((arg_dims := type_info.extract_dims(arg_type))).issubset(set(broadcast_dims)):
            raise errors.DSLError(
                node.location,
                f"Incompatible broadcast dimensions in '{node.func!s}': expected "
                f"broadcast dimension(s) '{set(arg_dims).difference(set(broadcast_dims))}' missing",
            )

        return_type = ts.FieldType(dims=broadcast_dims, dtype=type_info.extract_dtype(arg_type))

        return foast.Call(
            func=node.func,
            args=node.args,
            kwargs=node.kwargs,
            location=node.location,
            type=return_type,
        )

    def visit_Constant(self, node: foast.Constant, **kwargs: Any) -> foast.Constant:
        try:
            type_ = type_translation.from_value(node.value)
        except ValueError as e:
            raise errors.DSLError(node.location, "Could not deduce type of constant.") from e
        return foast.Constant(value=node.value, location=node.location, type=type_)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Transformation and optimization pipeline for the DaCe backend in GT4Py.

Please also see [ADR0018](https://github.com/GridTools/gt4py/tree/main/docs/development/ADRs/0018-Canonical_SDFG_in_GT4Py_Transformations.md)
that explains the general structure and requirements on the SDFGs.
"""

from . import constants, splitting_tools
from .auto_optimize import (
    GT4PyAutoOptHook,
    GT4PyAutoOptHookFun,
    GT4PyAutoOptHookStage,
    gt_auto_optimize,
)
from .concat_where_mapper import (
    gt_apply_concat_where_replacement_on_sdfg,
    gt_check_if_concat_where_node_is_replaceable,
    gt_replace_concat_where_node,
)
from .dead_dataflow_elimination import gt_eliminate_dead_dataflow, gt_remove_map
from .fuse_horizontal_conditionblocks import FuseHorizontalConditionBlocks
from .gpu_utils import (
    GPUSetBlockSize,
    gt_gpu_transform_non_standard_memlet,
    gt_gpu_transformation,
    gt_set_gpu_blocksize,
)
from .inline_fuser import inline_dataflow_into_map
from .local_double_buffering import gt_create_local_double_buffering
from .loop_blocking import LoopBlocking
from .map_fusion import (
    HorizontalMapFusionCallback,
    MapFusionHorizontal,
    MapFusionVertical,
    VerticalMapFusionCallback,
)
from .map_fusion_extended import (
    HorizontalMapSplitCallback,
    VerticalMapSplitCallback,
    gt_horizontal_map_split_fusion,
    gt_vertical_map_split_fusion,
)
from .map_orderer import MapIterationOrder, gt_set_iteration_order
from .map_promoter import MapPromoter, MapPromotionCallBack
from .map_splitter import MapSplitter
from .map_to_copy import MapToCopy
from .move_dataflow_into_if_body import MoveDataflowIntoIfBody
from .multi_state_global_self_copy_elimination import (
    MultiStateGlobalSelfCopyElimination,
    MultiStateGlobalSelfCopyElimination2,
    gt_multi_state_global_self_copy_elimination,
)
from .redundant_array_removers import CopyChainRemover, DoubleWriteRemover, gt_remove_copy_chain
from .remove_access_node_copies import RemoveAccessNodeCopies
from .remove_scalar_copies import RemoveScalarCopies
from .remove_views import RemovePointwiseViews
from .scan_loop_unrolling import ScanLoopUnrolling
from .simplify import (
    GT4PyMapBufferElimination,
    GT4PyMoveTaskletIntoMap,
    gt_inline_nested_sdfg,
    gt_reduce_distributed_buffering,
    gt_simplify,
    gt_substitute_compiletime_symbols,
)
from .single_state_global_self_copy_elimination import (
    SingleStateGlobalDirectSelfCopyElimination,
    SingleStateGlobalSelfCopyElimination,
)
from .split_access_nodes import SplitAccessNode, gt_split_access_nodes
from .split_memlet import SplitConsumerMemlet
from .state_fusion import GT4PyStateFusion
from .strides import (
    gt_change_strides,
    gt_map_strides_to_dst_nested_sdfg,
    gt_map_strides_to_src_nested_sdfg,
    gt_propagate_strides_from_access_node,
    gt_propagate_strides_of,
)
from .utils import gt_make_transients_persistent, unique_name


__all__ = [
    "CopyChainRemover",
    "DoubleWriteRemover",
    "FuseHorizontalConditionBlocks",
    "GPUSetBlockSize",
    "GT4PyAutoOptHook",
    "GT4PyAutoOptHookFun",
    "GT4PyAutoOptHookStage",
    "GT4PyMapBufferElimination",
    "GT4PyMoveTaskletIntoMap",
    "GT4PyStateFusion",
    "HorizontalMapFusionCallback",
    "HorizontalMapSplitCallback",
    "LoopBlocking",
    "MapFusionHorizontal",
    "MapFusionVertical",
    "MapIterationOrder",
    "MapPromoter",
    "MapPromotionCallBack",
    "MapSplitter",
    "MapToCopy",
    "MoveDataflowIntoIfBody",
    "MultiStateGlobalSelfCopyElimination",
    "MultiStateGlobalSelfCopyElimination2",
    "RemoveAccessNodeCopies",
    "RemovePointwiseViews",
    "RemoveScalarCopies",
    "ScanLoopUnrolling",
    "SingleStateGlobalDirectSelfCopyElimination",
    "SingleStateGlobalSelfCopyElimination",
    "SplitAccessNode",
    "SplitConsumerMemlet",
    "VerticalMapFusionCallback",
    "VerticalMapSplitCallback",
    "constants",
    "gt_apply_concat_where_replacement_on_sdfg",
    "gt_auto_optimize",
    "gt_change_strides",
    "gt_check_if_concat_where_node_is_replaceable",
    "gt_create_local_double_buffering",
    "gt_eliminate_dead_dataflow",
    "gt_gpu_transform_non_standard_memlet",
    "gt_gpu_transformation",
    "gt_horizontal_map_split_fusion",
    "gt_inline_nested_sdfg",
    "gt_make_transients_persistent",
    "gt_map_strides_to_dst_nested_sdfg",
    "gt_map_strides_to_src_nested_sdfg",
    "gt_multi_state_global_self_copy_elimination",
    "gt_propagate_strides_from_access_node",
    "gt_propagate_strides_of",
    "gt_reduce_distributed_buffering",
    "gt_remove_copy_chain",
    "gt_remove_map",
    "gt_replace_concat_where_node",
    "gt_set_gpu_blocksize",
    "gt_set_iteration_order",
    "gt_simplify",
    "gt_split_access_nodes",
    "gt_substitute_compiletime_symbols",
    "gt_vertical_map_split_fusion",
    "inline_dataflow_into_map",
    "splitting_tools",
    "unique_name",
]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

# TODO(iomaganaris): This transformation is very specific to a pattern in the vertically_implicit_solver of ICON4Py.
# The transformation should be generalized or moved to ICON4Py.
# One idea to generalize this transformation is: in case we have a chain of 4 AccessNodes which are
# connected sequentially with memlets (first_node -> second_node -> third_node -> fourth_node),
# we can split the middle access nodes (second and third) based on their input and output memlets' subsets and those can later
# be removed because the same data are copied from the first all the way to fourth node.
# The split should be done using SplitAccessNode transformation.

import warnings
from typing import Any, Optional

import dace
from dace import (
    data as dace_data,
    properties as dace_properties,
    transformation as dace_transformation,
)
from dace.sdfg import nodes as dace_nodes
from dace.transformation.passes import analysis as dace_analysis

from gt4py.next.program_processors.runners.dace.transformations import (
    splitting_tools as gtx_dace_split,
    strides as gtx_transformations_strides,
    utils as gtx_transformations_utils,
)


@dace_properties.make_properties
class RemoveAccessNodeCopies(dace_transformation.SingleStateTransformation):
    """
    Replace temporary AccessNodes' data with global AccessNode data in a pattern
    found in ICON4Py's dycore vertically_implicit_solver.
    The pattern is of the form:
        first_node (global) -> second_node (transient) -> third_node (transient) -> fourth_node (global)
    where first_node and fourth_node refer to the same data.
    The data of second_node and third_node are replaced with the data of first_node/fourth_node,
    and the subsets of the edges are updated accordingly.
    This transformation helps to reduce the number of data copies in the SDFG."""

    first_node = dace_transformation.PatternNode(dace_nodes.AccessNode)
    second_node = dace_transformation.PatternNode(dace_nodes.AccessNode)
    third_node = dace_transformation.PatternNode(dace_nodes.AccessNode)
    fourth_node = dace_transformation.PatternNode(dace_nodes.AccessNode)

    # Name of all data that is used at only one place. Is computed by the
    #  `FindSingleUseData` pass and be passed at construction time. Needed until
    #  [issue#1911](https://github.com/spcl/dace/issues/1911) has been solved.
    _single_use_data: Optional[dict[dace.SDFG, set[str]]]

    def __init__(
        self,
        *args: Any,
        single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._single_use_data = single_use_data

    @classmethod
    def expressions(cls) -> Any:
        return [
            dace.sdfg.utils.node_path_graph(
                cls.first_node, cls.second_node, cls.third_node, cls.fourth_node
            )
        ]

    def can_be_applied(
        self,
        graph: dace.SDFGState,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        first_node: dace_nodes.AccessNode = self.first_node
        first_desc: dace_data.Data = first_node.desc(sdfg)
        second_node: dace_nodes.AccessNode = self.second_node
        second_desc: dace_data.Data = second_node.desc(sdfg)
        third_node: dace_nodes.AccessNode = self.third_node
        third_desc: dace_data.Data = third_node.desc(sdfg)
        fourth_node: dace_nodes.AccessNode = self.fourth_node
        fourth_desc: dace_data.Data = fourth_node.desc(sdfg)

        # Match the pattern found in "vertically_implicit_solver"
        if not (
            first_desc.transient is False
            and second_desc.transient is True
            and third_desc.transient is True
            and fourth_desc.transient is False
        ):
            return False

        scope_dict = graph.scope_dict()
        # Make sure that all the access nodes node are not in any scope
        if scope_dict[first_node] is not None:
            return False
        if scope_dict[second_node] is not None:
            return False
        if scope_dict[third_node] is not None:
            return False
        if scope_dict[fourth_node] is not None:
            return False

        # Make sure that first and fourth node refer to the same data
        if first_node.data != fourth_node.data:
            return False

        # Make sure that second and third node have the same shape
        if second_desc.shape != third_desc.shape:
            return False

        # Make sure that second and third node are not used anywhere else in the SDFG
        if self._single_use_data is None:
            find_single_use_data = dace_analysis.FindSingleUseData()
            single_use_data = find_single_use_data.apply_pass(sdfg, None)
        else:
            single_use_data = self._single_use_data
        if (
            second_node.data not in single_use_data[sdfg]
            and third_node.data not in single_use_data[sdfg]
        ):
            return False

        # Make sure that there is no other AccessNode with the same data as first and fourth nodes in the SDFG state
        for node in graph.data_nodes():
            if node not in [first_node, fourth_node] and node.data == first_node.data:
                return False

        # Make sure that the data written to the first node are not a subset of the data written to the fourth node
        ranges_written_to_fourth_node = []
        for edge in graph.in_edges(fourth_node):
            ranges_written_to_fourth_node.append(edge.data.dst_subset)

        ranges_written_to_first_node = []
        for edge in graph.in_edges(first_node):
            ranges_written_to_first_node.append(edge.data.dst_subset)
            if any(
                gtx_dace_split.are_intersecting(edge.data.dst_subset, fourth_node_range)
                for fourth_node_range in ranges_written_to_fourth_node
            ):
                return False

        # Make sure that the whole data of the first node is written
        first_node_range = first_desc.shape

        first_node_writes_subset = gtx_dace_split.subset_merger(ranges_written_to_first_node)
        if len(first_node_writes_subset) != 1:
            warnings.warn(
                "[RemoveAccessNodeCopies] The range of writes to the first node is not a single range.",
                stacklevel=0,
            )
            return False
        fourth_node_writes_subset = gtx_dace_split.subset_merger(ranges_written_to_fourth_node)
        if len(fourth_node_writes_subset) != 1:
            warnings.warn(
                "[RemoveAccessNodeCopies] The range of writes to the fourth node is not a single range.",
                stacklevel=0,
            )
            return False
        union_written_to_first_node_data = next(iter(first_node_writes_subset))
        union_written_to_fourth_node_data = next(iter(fourth_node_writes_subset))
        union_written_to_common_data = gtx_dace_split.subset_merger(
            [union_written_to_first_node_data, union_written_to_fourth_node_data]
        )
        if len(union_written_to_common_data) != 1:
            warnings.warn(
                "[RemoveAccessNodeCopies] The union of the ranges written to the first and fourth nodes is not a single range.",
                stacklevel=0,
            )
            return False
        if union_written_to_common_data != first_node_range:
            warnings.warn(
                "[RemoveAccessNodeCopies] The whole range of the first node is not written.",
                stacklevel=0,
            )
            # When https://github.com/GridTools/gt4py/pull/2173 is merged we can return False
            # At the moment we don't have the range of the global AccessNodes at compile time

        # Make sure that data written to first node are only copied from first to second node
        if any(edge.dst != second_node for edge in graph.out_edges(first_node)):
            return False

        # Make sure that there are no edges from second node that end up in any node apart from the third node
        for edge in graph.out_edges(second_node):
            if edge.dst != third_node and not gtx_transformations_utils.is_reachable(
                start=edge.dst,
                target=third_node,
                state=graph,
            ):
                return False
        source_nodes_of_third_node = gtx_transformations_utils.find_upstream_nodes(
            start=third_node,
            state=graph,
        )
        if second_node not in source_nodes_of_third_node:
            return False

        return True

    def apply(
        self,
        graph: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> None:
        first_node: dace_nodes.AccessNode = self.first_node
        second_node: dace_nodes.AccessNode = self.second_node
        third_node: dace_nodes.AccessNode = self.third_node

        # Find the edge between first and second access nodes (there should be only one)
        edge_between_first_and_second = next(
            edge for edge in graph.edges_between(first_node, second_node)
        )

        # Compute the offset of the data between first and second node
        assert (
            edge_between_first_and_second.data.src_subset.dims()
            == edge_between_first_and_second.data.dst_subset.dims()
        )
        node_offset = [
            range_first_min - range_second_min
            for range_first_min, range_second_min in zip(
                edge_between_first_and_second.data.src_subset.min_element(),
                edge_between_first_and_second.data.dst_subset.min_element(),
            )
        ]

        # Make sure that the `subset` and `other_subset` of all memlets inside the map scopes
        # of the SDFG state are referring to the data of the access nodes which are outside the map scopes
        scope_dict = graph.scope_dict()
        for node in graph.nodes():
            if isinstance(node, (dace_nodes.EntryNode)) and scope_dict[node] is None:
                dace.sdfg.utils.canonicalize_memlet_trees_for_map(graph, node)
                dace.sdfg.utils.canonicalize_memlet_trees_for_map(graph, graph.exit_node(node))

        # Update all edges with the correct subsets that match the first node range
        for edge in graph.edges():
            # Update edges that refer to second or third AccessNodes
            if edge.data.data == second_node.data or edge.data.data == third_node.data:
                # Replace data with data of first node
                edge.data.data = first_node.data
                # Add offset to subset
                edge.data.subset.offset(node_offset, negative=False)
                # Add offset to other_subset if it's necessary (src is either first, second or third node and dst is either second or third node)
                if edge.dst in [second_node, third_node] and edge.data.other_subset is not None:
                    edge.data.other_subset.offset(node_offset, negative=False)
            elif edge.data.data == first_node.data and edge.dst in [second_node, third_node]:
                # Edge from first node to second or third node
                # Update other_subset to match the subset of the first node
                edge.data.other_subset.offset(node_offset, negative=False)

        # Replace data of second and third node with data of first and fourth node
        second_node.data = first_node.data
        third_node.data = first_node.data

        # Update strides after changing the data of access nodes
        for modified_node in [second_node, third_node]:
            gtx_transformations_strides.gt_propagate_strides_from_access_node(
                sdfg, graph, modified_node
            )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import collections
import concurrent.futures
import contextlib
import contextvars
import dataclasses
import functools
import itertools
import threading
import warnings
import weakref
from collections.abc import Callable, Hashable, Sequence
from typing import Any, Generic, TypeAlias, TypeVar

from gt4py._core import definitions as core_defs
from gt4py.eve import extended_typing as xtyping, utils as eve_utils
from gt4py.next import backend as gtx_backend, common, config, errors, utils as gtx_utils
from gt4py.next.ffront import (
    stages as ffront_stages,
    type_info as ffront_type_info,
    type_specifications as ts_ffront,
    type_translation,
)
from gt4py.next.instrumentation import hook_machinery, metrics
from gt4py.next.otf import arguments, stages
from gt4py.next.type_system import type_specifications as ts
from gt4py.next.utils import tree_map


T = TypeVar("T")

ScalarOrTupleOfScalars: TypeAlias = xtyping.MaybeNestedInTuple[core_defs.Scalar]

#: Content of the key: (*hashable_arg_descriptors, id(offset_provider), concrete_instantation_if_generic)
CompiledProgramsKey: TypeAlias = tuple[tuple[Hashable, ...], int, None | str]

ArgStaticDescriptorsByType: TypeAlias = dict[
    type[arguments.ArgStaticDescriptor], dict[str, arguments.ArgStaticDescriptor]
]

#: Counter that uniquely identifies every compiled program pool created from the same root.
#: It is only used to give a meaningful name to the metrics source key.
_pools_per_root: collections.Counter = collections.Counter()
_pools_per_root_lock: threading.Lock = threading.Lock()

# Cache metrics source keys for each compiled program pool id and key.
# Note: we use a weakref finalizer to remove entries from this cache when a
# compiled program pool is deleted, to avoid creating false cache entries
# since the id of the program pool could be reused by another pool once the
# original one has been deleted.
_metrics_source_key_cache: contextvars.ContextVar[dict[tuple[int, CompiledProgramsKey], str]] = (
    contextvars.ContextVar("_metrics_source_key_cache")
)


def metrics_source_key(pool: CompiledProgramsPool, key: CompiledProgramsKey) -> str:
    """Generate a metrics source key from a concrete item of a compiled programs pool."""
    _metrics_key_cache = _metrics_source_key_cache.get(None)
    if _metrics_key_cache is None:
        _metrics_key_cache = {}
        _metrics_source_key_cache.set(_metrics_key_cache)
    try:
        return _metrics_key_cache[(id(pool), key)]
    except KeyError:
        source_key = f"{pool.root[0]}<{pool.root[1]}>#{_pools_per_root[pool.root]}[{hash(key)}]"
        source_key_cache_entry = (id(pool), key)
        _metrics_key_cache[source_key_cache_entry] = source_key
        # Use a finalizer to remove the entry from the cache once the pool is deleted
        weakref.finalize(
            pool,
            lambda entry, cache: cache.pop(entry, None),
            source_key_cache_entry,
            _metrics_key_cache,
        )
        return source_key


@hook_machinery.event_hook
def compile_variant_hook(
    program_pool: CompiledProgramsPool,
    key: CompiledProgramsKey,
    backend: gtx_backend.Backend,
    argument_descriptors: ArgStaticDescriptorsByType,
    offset_provider: common.OffsetProviderType | common.OffsetProvider,
) -> None:
    """Callback hook invoked before compiling a program variant."""

    if metrics.is_any_level_enabled():
        # Create a new metrics entity for this compiled program variant and
        # attach relevant metadata to it.
        source_key = metrics_source_key(program_pool, key)
        assert source_key not in metrics.sources, (
            "The key for the program variant being compiled is already set!!"
        )

        metrics.sources[source_key].metadata |= dict(
            name=program_pool.definition_stage.definition.__name__,
            backend=backend.name,
            compiled_program_pool_id=id(program_pool),
            compiled_program_pool_key=hash(key),
            **{
                f"{eve_utils.CaseStyleConverter.convert(key.__name__, 'pascal', 'snake')}s": value
                for key, value in argument_descriptors.items()
            },
        )

    if __debug__:
        # Note: We set the stack level to point to something internally as we don't want to show this warning for every program call.
        # It's an ad-hoc pragmatic choice that could be revisited in the future.
        warnings.warn(
            "Python is not running in optimized mode, which may impact performance."
            " Consider running with `python -O` or setting the environment"
            " variable `PYTHONOPTIMIZE=1`.",
            stacklevel=3,
        )


@hook_machinery.context_hook
def compiled_program_call_context(
    program_pool: CompiledProgramsPool,
    key: CompiledProgramsKey,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
    offset_provider: common.OffsetProvider,
) -> contextlib.AbstractContextManager:
    """
    Hook called at the beginning and end of a compiled program call.

    Args:
        compiled_program: The compiled program being called.
        args: The arguments with which the program is called.
        kwargs: The keyword arguments with which the program is called.
        offset_provider: The offset provider passed to the program.
        root: The root of the compiled programs pool this program belongs to, i.e. a tuple of
            (program name, backend name).
        key: The key of the compiled program in the compiled programs pool.

    """
    # We set the metrics key for the compiled program call at enter and leave it
    # set at exit, since it is needed in the exit of the outer `program_call` context,
    # which will take care of unsetting it. This is because the compiled program call
    # is part of a program call, but we want the metrics to be associated with a
    # specific compiled program variant, not just the generic outer program.
    return metrics.metrics_source_key_setter(metrics_source_key(program_pool, key))


# TODO(havogt): We would like this to be a ProcessPoolExecutor, which requires (to decide what) to pickle.
_async_compilation_pool: concurrent.futures.Executor | None = None


def _init_async_compilation_pool() -> None:
    global _async_compilation_pool
    if _async_compilation_pool is None and config.BUILD_JOBS > 0:
        _async_compilation_pool = concurrent.futures.ThreadPoolExecutor(
            max_workers=config.BUILD_JOBS
        )


_init_async_compilation_pool()


def wait_for_compilation() -> None:
    """
    Waits for all ongoing compilations to finish.

    This is useful to ensure that all compiled programs are ready before
    proceeding with further operations. E.g. when the first call is included in timings.
    """
    global _async_compilation_pool
    if _async_compilation_pool is not None:
        _async_compilation_pool.shutdown(wait=True)
        _async_compilation_pool = None
        _init_async_compilation_pool()


def _make_tuple_expr(el_exprs: list[str]) -> str:
    return "".join((f"{el},") for el in el_exprs)


def _make_param_context_from_func_type(
    func_type: ts.FunctionType,
    type_map: Callable[[ts.TypeSpec], T] = lambda x: x,  # type: ignore[assignment, return-value]  # mypy not smart enough to narrow type for default
) -> dict[str, xtyping.MaybeNestedInTuple[T]]:
    """
    Create a context to evaluate expressions in from a function type.

    >>> int32_t, int64_t = (
    ...     ts.ScalarType(kind=ts.ScalarKind.INT32),
    ...     ts.ScalarType(kind=ts.ScalarKind.INT64),
    ... )
    >>> type_ = ts.FunctionType(
    ...     pos_only_args=[],
    ...     pos_or_kw_args={"inp1": ts.TupleType(types=[int32_t, int64_t])},
    ...     kw_only_args={"inp2": int64_t},
    ...     returns=int64_t,
    ... )
    >>> context = _make_param_context_from_func_type(type_)
    >>> assert context == {"inp1": (int32_t, int64_t), "inp2": int64_t}
    """
    params = func_type.pos_or_kw_args | func_type.kw_only_args
    return {
        param: ffront_type_info.tree_map_type(
            # note(tehrengruber): Mapping all collections to tuples is and must be the same as in
            #  :ref:`arguments.extract`.
            type_map,
            result_collection_constructor=lambda _, els: tuple(els),
        )(type_)
        for param, type_ in params.items()
    }


def _get_type_of_param_expr(program_type: ts_ffront.ProgramType, expr: str) -> ts.TypeSpec:
    structured_type_ = eval(expr, _make_param_context_from_func_type(program_type.definition))
    type_ = tree_map(
        lambda v: v, result_collection_constructor=lambda _, elts: ts.TupleType(types=list(elts))
    )(structured_type_)
    assert isinstance(type_, ts.TypeSpec)
    return type_


def _make_argument_descriptors(
    program_type: ts_ffront.ProgramType,
    argument_descriptor_mapping: dict[type[arguments.ArgStaticDescriptor], Sequence[str]],
    args: tuple[Any],
    kwargs: dict[str, Any],
) -> ArgStaticDescriptorsByType:
    """Given a set of runtime arguments construct all argument descriptors from them."""
    func_type = program_type.definition
    params = list(func_type.pos_or_kw_args.keys()) + list(func_type.kw_only_args.keys())
    descriptors: ArgStaticDescriptorsByType = {}
    for descriptor_cls, exprs in argument_descriptor_mapping.items():
        descriptors[descriptor_cls] = {}
        for expr in exprs:
            argument = eval(f"""lambda {",".join(params)}: {expr}""")(*args, **kwargs)
            descriptors[descriptor_cls][expr] = descriptor_cls.from_value(argument)
    _validate_argument_descriptors(program_type, descriptors)
    return descriptors


def _convert_to_argument_descriptor_context(
    func_type: ts.FunctionType, argument_descriptors: ArgStaticDescriptorsByType
) -> arguments.ArgStaticDescriptorsContextsByType:
    """
    Given argument descriptors, i.e., a mapping from an expr to a descriptor, transform them into a
    context of argument descriptors in which we can evaluate expressions.

    >>> int32_t, int64_t = (
    ...     ts.ScalarType(kind=ts.ScalarKind.INT32),
    ...     ts.ScalarType(kind=ts.ScalarKind.INT64),
    ... )
    >>> type_ = ts.FunctionType(
    ...     pos_only_args=[],
    ...     pos_or_kw_args={"inp1": ts.TupleType(types=[int32_t, int64_t])},
    ...     kw_only_args={"inp2": int64_t},
    ...     returns=int64_t,
    ... )
    >>> argument_descriptors = {arguments.StaticArg: {"inp1[1]": arguments.StaticArg(value=1)}}
    >>> contexts = _convert_to_argument_descriptor_context(type_, argument_descriptors)
    >>> contexts[arguments.StaticArg]
    {'inp1': (None, StaticArg(value=1)), 'inp2': None}
    """
    descriptor_contexts: arguments.ArgStaticDescriptorsContextsByType = {}
    for descriptor_cls, descriptor_expr_mapping in argument_descriptors.items():
        context: arguments.ArgStaticDescriptorsContext = _make_param_context_from_func_type(
            func_type, lambda x: None
        )
        # convert tuples to list such that we can alter the context easily
        context = {
            k: gtx_utils.tree_map(
                lambda v: v,
                collection_type=tuple,
                result_collection_constructor=lambda _, elts: list(elts),
            )(v)
            for k, v in context.items()
        }
        assert "__descriptor" not in context
        for expr, descriptor in descriptor_expr_mapping.items():
            # note: we don't need to handle any errors here since the `expr` has been validated
            #  in `_validate_argument_descriptor_mapping`
            exec(
                f"{expr} = __descriptor",
                {"__descriptor": descriptor},
                context,
            )
        # convert lists back to tuples
        context = {
            k: gtx_utils.tree_map(
                lambda v: v,
                collection_type=list,
                result_collection_constructor=lambda _, elts: tuple(elts),
            )(v)
            for k, v in context.items()
        }
        descriptor_contexts[descriptor_cls] = context  # type: ignore[index]  # Hard to understand, it looks like a mypy bug

    return descriptor_contexts


def _validate_argument_descriptors(
    program_type: ts_ffront.ProgramType, all_descriptors: ArgStaticDescriptorsByType
) -> None:
    for descriptors in all_descriptors.values():
        for expr, descriptor in descriptors.items():
            param_type = _get_type_of_param_expr(program_type, expr)
            descriptor.validate(expr, param_type)


@dataclasses.dataclass
class CompiledProgramsPool(Generic[ffront_stages.DSLDefinitionT]):
    """
    A pool of compiled programs for a given program and backend.

    If 'argument_descriptor_mapping' is populated the pool will create a program for each
    argument that has an argument descriptor. E.g., if a param is marked static we create
    a new program for each value of that parameter. See :ref:`arguments.ArgumentDescriptor` for
    more information on argument descriptors.

    If `enable_jit` is True in the call to the pool, it will compile a program
    with static information as described in `argument_descriptor_mapping`, otherwise it
    will error. In the latter case, the pool needs to be filled with call(s)
    to `compile` before it can be used.
    """

    backend: gtx_backend.Backend
    definition_stage: ffront_stages.DSLDefinitionT
    # Note: This type can be incomplete, i.e. contain DeferredType, whenever the operator is a
    #  scan operator. In the future it could also be the type of a generic program.
    program_type: ts_ffront.ProgramType
    #: mapping from an argument descriptor type to a list of parameters or expression thereof
    #: e.g. `{arguments.StaticArg: ["static_int_param"]}`
    #: Note: The list is not ordered.
    argument_descriptor_mapping: dict[type[arguments.ArgStaticDescriptor], Sequence[str]] | None

    # store for the compiled programs
    compiled_programs: dict[CompiledProgramsKey, stages.ExecutableProgram] = dataclasses.field(
        default_factory=dict, init=False
    )

    # store for the async compilation jobs
    _compilation_jobs: dict[
        CompiledProgramsKey, concurrent.futures.Future[stages.ExecutableProgram]
    ] = dataclasses.field(default_factory=dict, init=False)

    @functools.cached_property
    def root(self) -> tuple[str, str]:
        result = (self.definition_stage.definition.__name__, self.backend.name)
        with _pools_per_root_lock:
            _pools_per_root[result] += 1
        return result

    def __post_init__(self) -> None:
        # TODO(havogt): We currently don't support pos_only or kw_only args at the program level.
        # This check makes sure we don't miss updating this code if we add support for them in the future.
        assert not self.program_type.definition.kw_only_args
        assert not self.program_type.definition.pos_only_args
        self._validate_argument_descriptor_mapping()

        # Force initialization of all cached properties here to minimize first-time call overhead
        self._primitive_values_extractor  # noqa: B018

    def __call__(
        self, *args: Any, offset_provider: common.OffsetProvider, enable_jit: bool, **kwargs: Any
    ) -> None:
        """
        Calls a program with the given arguments and offset provider.

        If the program is not in cache, it will jit compile with static arguments
        (defined by 'static_params') in case `enable_jit` is True. Otherwise,
        it is an error.
        """
        canonical_args, canonical_kwargs = self._args_canonicalizer(*args, **kwargs)
        if (extractor := self._primitive_values_extractor) is not None:
            args, kwargs = extractor(*canonical_args, **canonical_kwargs)
        else:
            args, kwargs = canonical_args, canonical_kwargs
        static_args_values = self._argument_descriptor_cache_key_from_args(*args, **kwargs)

        if self._is_generic:
            # In case the program or operator is generic, i.e. callable for arguments of varying
            # type, add the argument types to the cache key as the argument types are used during
            # compilation. In case the program is not generic we can avoid the potentially
            # expensive type deduction for all arguments and not include it in the key.
            if enable_jit:
                warnings.warn(
                    "Calling generic programs / direct calls to scan operators are not optimized. "
                    "Consider calling a specialized version instead.",
                    stacklevel=3,
                )
            arg_specialization_key = eve_utils.content_hash(
                (
                    tuple(type_translation.from_value(arg) for arg in canonical_args),
                    {k: type_translation.from_value(v) for k, v in canonical_kwargs.items()},
                )
            )
        else:
            arg_specialization_key = None

        key = (
            static_args_values,
            common.hash_offset_provider_items_by_id(offset_provider),
            arg_specialization_key,
        )

        try:
            compiled_program = self.compiled_programs[key]

        except KeyError as e:
            if self._finish_compilation_job(key):
                compiled_program = self.compiled_programs[key]
            elif enable_jit:
                assert self.argument_descriptor_mapping is not None
                self._compile_variant(
                    argument_descriptors=_make_argument_descriptors(
                        self.program_type, self.argument_descriptor_mapping, args, kwargs
                    ),
                    # note: it is important to use the args before named collections are extracted
                    #  as otherwise the implicit program generation from an operator fails
                    arg_specialization_info=(
                        tuple(type_translation.from_value(arg) for arg in canonical_args),
                        {k: type_translation.from_value(v) for k, v in canonical_kwargs.items()},
                    ),
                    offset_provider=offset_provider,
                    call_key=key,
                )
                return self(
                    *canonical_args,
                    offset_provider=offset_provider,
                    enable_jit=False,
                    **canonical_kwargs,
                )  # passing `enable_jit=False` because a cache miss should be a hard-error in this call`

            else:
                raise RuntimeError("No program compiled for this set of static arguments.") from e

        with compiled_program_call_context(self, key, args, kwargs, offset_provider):
            compiled_program(*args, **kwargs, offset_provider=offset_provider)

    @functools.cached_property
    def _primitive_values_extractor(self) -> Callable | None:
        return arguments.make_primitive_value_args_extractor(self.program_type.definition)

    @functools.cached_property
    def _is_generic(self) -> bool:
        """
        Is the operator or program generic in the sense that it can be called for different
        argument types.

        Right now this is only the case for scan operators.
        """
        # TODO(tehrengruber): This concept does not exist elsewhere and is not properly reflected
        #  in the type system. For now we just use `DeferredType` to communicate between
        #  here and `type_info.type_in_program_context`.
        return any(
            isinstance(t, ts.DeferredType)
            for t in itertools.chain(
                self.program_type.definition.pos_only_args,
                self.program_type.definition.pos_or_kw_args.values(),
                self.program_type.definition.kw_only_args.values(),
            )
        )

    @functools.cached_property
    def _args_canonicalizer(self) -> Callable[..., tuple[tuple, dict[str, Any]]]:
        return ffront_type_info.make_args_canonicalizer(
            self.program_type, name=self.definition_stage.definition.__name__
        )

    @functools.cached_property
    def _argument_descriptor_cache_key_from_args(
        self,
    ) -> Callable[..., tuple[Hashable, ...]]:
        """
        Given the entire set of runtime arguments compute the cache key used to retrieve the
        instance of the compiled program which is compiled for the argument descriptors from
        the given set of arguments.

        This is part of the performance critical path that is called on every program call,
        hence we code generate a single lambda expression here.
        """
        func_type = self.program_type.definition
        params = list(func_type.pos_or_kw_args.keys()) + list(func_type.kw_only_args.keys())
        elements: list[str] = []
        for descriptor_cls, arg_exprs in self.argument_descriptor_mapping.items():  # type: ignore[union-attr]  # can never be `None` at this point
            for arg_expr in arg_exprs:
                attr_extractor = descriptor_cls.attribute_extractor_exprs(arg_expr)
                elements.extend(attr_extractor.values())
        return eval(f"""lambda {",".join(params)}: ({_make_tuple_expr(elements)})""")

    def _argument_descriptor_cache_key_from_descriptors(
        self,
        argument_descriptor_contexts: arguments.ArgStaticDescriptorsContextsByType,
    ) -> tuple:
        """
        Given a set of argument descriptors deduce the cache key used to retrieve the instance
        of the compiled program which is compiled for the given argument descriptors.

        This function is not performance critical as it is only called once when compiling a
        variant.
        """
        elements = []
        for descriptor_cls, arg_exprs in self.argument_descriptor_mapping.items():  # type: ignore[union-attr]  # can never be `None` at this point
            for arg_expr in arg_exprs:
                attr_extractor = descriptor_cls.attribute_extractor_exprs(arg_expr)
                attrs = attr_extractor.keys()
                for attr in attrs:
                    elements.append(
                        getattr(
                            eval(f"{arg_expr}", {}, argument_descriptor_contexts[descriptor_cls]),
                            attr,
                        )
                    )
        return tuple(elements)

    def _initialize_argument_descriptor_mapping(
        self, argument_descriptors: ArgStaticDescriptorsByType
    ) -> None:
        if self.argument_descriptor_mapping is None:
            self.argument_descriptor_mapping = {
                descr_cls: list(descriptor_expr_mapping.keys())
                for descr_cls, descriptor_expr_mapping in argument_descriptors.items()
            }
            self._validate_argument_descriptor_mapping()
        else:
            for descr_cls, descriptor_expr_mapping in argument_descriptors.items():
                if (expected := set(self.argument_descriptor_mapping.get(descr_cls, {}))) != (
                    got := set(descriptor_expr_mapping.keys())
                ):
                    raise ValueError(
                        f"Argument descriptor {descr_cls.__name__} must be the same for all compiled programs, got {list(got)} expected {list(expected)}."
                    )

    def _validate_argument_descriptor_mapping(self) -> None:
        if self.argument_descriptor_mapping is None:
            return
        context = _make_param_context_from_func_type(self.program_type.definition, lambda x: None)
        for descr_cls, exprs in self.argument_descriptor_mapping.items():
            for expr in exprs:
                try:
                    if any(
                        v is not None for v in gtx_utils.flatten_nested_tuple(eval(expr, context))
                    ):
                        raise ValueError()
                except (ValueError, KeyError, NameError):
                    raise errors.DSLTypeError(  # noqa: B904 # we don't care about the original exception
                        message=f"Invalid parameter expression '{expr}' for '{descr_cls.__name__}'. "
                        f"Must be the name of a parameter or an access to one of its elements.",
                        location=None,
                    )

    def _is_existing_key(self, key: CompiledProgramsKey) -> bool:
        return key in self.compiled_programs or key in self._compilation_jobs

    def _finish_compilation_job(self, key: CompiledProgramsKey) -> bool:
        if key not in self._compilation_jobs:
            return False

        compiled_program_future = self._compilation_jobs.pop(key)
        assert isinstance(compiled_program_future, concurrent.futures.Future)
        assert key not in self.compiled_programs
        self.compiled_programs[key] = compiled_program_future.result()
        return True

    def _compile_variant(
        self,
        argument_descriptors: ArgStaticDescriptorsByType,
        offset_provider: common.OffsetProviderType | common.OffsetProvider,
        #: tuple consisting of the types of the positional and keyword arguments.
        arg_specialization_info: tuple[tuple[ts.TypeSpec, ...], dict[str, ts.TypeSpec]]
        | None = None,
        # argument used only to validate key computed in a call / dispatch agrees with the
        # key computed here
        call_key: CompiledProgramsKey | None = None,
    ) -> None:
        if not common.is_offset_provider(offset_provider):
            if common.is_offset_provider_type(offset_provider):
                raise ValueError(
                    "Variant compilation of programs with 'OffsetProviderType' is not yet supported."
                )
            else:
                raise ValueError(f"Invalid 'offset_provider': {offset_provider}")

        self._initialize_argument_descriptor_mapping(argument_descriptors)
        _validate_argument_descriptors(self.program_type, argument_descriptors)

        argument_descriptor_contexts = _convert_to_argument_descriptor_context(
            self.program_type.definition, argument_descriptors
        )
        key = (
            self._argument_descriptor_cache_key_from_descriptors(argument_descriptor_contexts),
            common.hash_offset_provider_items_by_id(offset_provider),
            eve_utils.content_hash(arg_specialization_info) if self._is_generic else None,
        )
        assert call_key is None or call_key == key

        if self._is_existing_key(key):
            raise ValueError(f"Program with key {key} already exists.")

        if arg_specialization_info:
            arg_types, kwarg_types = arg_specialization_info
        else:
            if self._is_generic:
                raise ValueError(
                    "Can not precompile generic program or scan operator without argument types."
                )
            arg_types = (
                *self.program_type.definition.pos_only_args,
                *self.program_type.definition.pos_or_kw_args.values(),
            )
            kwarg_types = self.program_type.definition.kw_only_args

        compile_time_args = arguments.CompileTimeArgs(
            offset_provider=offset_provider,
            column_axis=None,  # TODO(havogt): column_axis seems to a unused, even for programs with scans
            args=arg_types,
            kwargs=kwarg_types,
            argument_descriptor_contexts=argument_descriptor_contexts,
        )
        compile_call = functools.partial(
            self.backend.compile, self.definition_stage, compile_time_args=compile_time_args
        )
        compile_variant_hook(
            self,
            key=key,
            backend=self.backend,
            argument_descriptors=argument_descriptors,
            offset_provider=offset_provider,
        )

        if _async_compilation_pool is None:
            self.compiled_programs[key] = compile_call()
        else:
            self._compilation_jobs[key] = _async_compilation_pool.submit(compile_call)

    # TODO(tehrengruber): Rework the interface to allow precompilation with compile time
    #  domains and of scans.
    def compile(
        self,
        offset_providers: list[common.OffsetProvider | common.OffsetProviderType],
        **static_args: list[ScalarOrTupleOfScalars],
    ) -> None:
        """
        Compiles the program for all combinations of static arguments and the given 'OffsetProviderType'.

        Note: In case you want to compile for specific combinations of static arguments (instead
        of the combinatoral), you can call compile multiples times.

        Examples:
            pool.compile(static_arg0=[0,1], static_arg1=[2,3], ...)
                will compile for (0,2), (0,3), (1,2), (1,3)
            pool.compile(static_arg0=[0], static_arg1=[2]).compile(static_arg=[1], static_arg1=[3])
                will compile for (0,2), (1,3)
        """
        for offset_provider in offset_providers:  # not included in product for better type checking
            for static_values in itertools.product(*static_args.values()):
                self._compile_variant(
                    argument_descriptors={
                        arguments.StaticArg: dict(
                            zip(
                                static_args.keys(),
                                [arguments.StaticArg(value=v) for v in static_values],
                                strict=True,
                            )
                        ),
                    },
                    offset_provider=offset_provider,
                )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses

import gt4py.next.type_system.type_specifications as ts
from gt4py.eve import codegen
from gt4py.next.otf import code_specs


def format_source(source_code_spec: code_specs.SourceCodeSpec, source: str) -> str:
    assert source_code_spec.formatter_key is not None, (
        "No formatter key specified in source code specification."
    )
    return codegen.format_source(
        source_code_spec.formatter_key, source, **(source_code_spec.formatter_options or {})
    )


@dataclasses.dataclass(frozen=True)
class Parameter:
    name: str
    type_: ts.TypeSpec


@dataclasses.dataclass(frozen=True)
class Function:
    name: str
    parameters: tuple[Parameter, ...]
    returns: bool = False


@dataclasses.dataclass(frozen=True)
class LibraryDependency:
    name: str
    version: str



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import warnings
from typing import Any, Final

import factory

import gt4py.next.custom_layout_allocators as next_allocators
from gt4py._core import definitions as core_defs
from gt4py.next import backend, common, config
from gt4py.next.otf import stages, workflow
from gt4py.next.program_processors.runners.dace.workflow.factory import DaCeWorkflowFactory


class DaCeBackendFactory(factory.Factory):
    """
    Workflow factory for the GTIR-DaCe backend.

    Several parameters are inherithed from `backend.Backend`, see below the specific ones.

    Args:
        auto_optimize: Enables the SDFG transformation pipeline.
    """

    class Meta:
        model = backend.Backend

    class Params:
        name_device = "cpu"
        name_cached = ""
        name_postfix = ""
        gpu = factory.Trait(
            allocator=next_allocators.StandardGPUFieldBufferAllocator(),
            device_type=core_defs.CUPY_DEVICE_TYPE or core_defs.DeviceType.CUDA,
            name_device="gpu",
        )
        cached = factory.Trait(
            executor=factory.LazyAttribute(
                lambda o: workflow.CachedStep(o.otf_workflow, hash_function=o.hash_function)
            ),
            name_cached="_cached",
        )
        device_type = core_defs.DeviceType.CPU
        hash_function = stages.compilation_hash
        otf_workflow = factory.SubFactory(
            DaCeWorkflowFactory,
            device_type=factory.SelfAttribute("..device_type"),
            auto_optimize=factory.SelfAttribute("..auto_optimize"),
        )
        auto_optimize = factory.Trait(name_postfix="_opt")

    name = factory.LazyAttribute(
        lambda o: f"run_dace_{o.name_device}{o.name_cached}{o.name_postfix}"
    )

    executor = factory.LazyAttribute(lambda o: o.otf_workflow)
    allocator = next_allocators.StandardCPUFieldBufferAllocator()
    transforms = backend.DEFAULT_TRANSFORMS


def make_dace_backend(
    gpu: bool,
    cached: bool = True,
    auto_optimize: bool = True,
    async_sdfg_call: bool = True,
    optimization_args: dict[str, Any] | None = None,
    unstructured_horizontal_has_unit_stride: bool = config.UNSTRUCTURED_HORIZONTAL_HAS_UNIT_STRIDE,
    use_metrics: bool = True,
    use_zero_origin: bool = False,
    use_max_domain_range_on_unstructured_shift: bool | None = None,
) -> backend.Backend:
    """Customize the dace backend with the given configuration parameters.

    Args:
        gpu: Enable GPU transformations and code generation.
        cached: Cache the lowered SDFG as a JSON file and the compiled programs.
        auto_optimize: Enable the SDFG auto-optimize pipeline.
        async_sdfg_call: Make an asynchronous SDFG call on GPU to allow overlapping
            of GPU kernel execution with the Python driver code.
        optimization_args: A `dict` containing configuration parameters for
            the SDFG auto-optimize pipeline, see `gt_auto_optimize()`.
        unstructured_horizontal_has_unit_stride: When the memory layout has unit stride
            in the horizontal dimension, replace the field stride symbol with '1'.
        use_metrics: Add SDFG instrumentation to collect the metric for stencil
            compute time.
        use_zero_origin: Can be set to `True` when all fields passed as program
            arguments have zero-based origin. This setting will skip generation
            of range start-symbols `_range_0` since they can be assumed to be zero.

    Note that `gt_auto_optimize()` parameters that are derived from GT4Py configuration
    cannot be overriden, and therefore cannot appear here. Thus, this function will
    throw an exception if called with any argument included in `gt_optimization_args`.

    Returns:
        A dace backend with custom configuration for the target device.
    """

    # The `gt_optimization_args` set contains the parameters of `gt_auto_optimize()`
    # that are derived from the gt4py configuration, and therefore cannot be customized.
    gt_optimization_args: Final[set[str]] = {"gpu", "constant_symbols", "unit_strides_kind"}

    if optimization_args is None:
        optimization_args = {}
    elif optimization_args and not auto_optimize:
        warnings.warn("Optimizations args given, but auto-optimize is disabled.", stacklevel=2)
    elif intersect_args := gt_optimization_args.intersection(optimization_args.keys()):
        raise ValueError(
            f"The following optimization arguments cannot be overriden: {intersect_args}."
        )

    # Set `unit_strides_kind` based on the gt4py env configuration.
    optimization_args = optimization_args | {
        "unit_strides_kind": common.DimensionKind.HORIZONTAL
        if unstructured_horizontal_has_unit_stride
        else None
    }

    return DaCeBackendFactory(  # type: ignore[return-value] # factory-boy typing not precise enough
        gpu=gpu,
        cached=cached,
        auto_optimize=auto_optimize,
        otf_workflow__cached_translation=cached,
        otf_workflow__bare_translation__async_sdfg_call=(async_sdfg_call if gpu else False),
        otf_workflow__bare_translation__auto_optimize_args=optimization_args,
        otf_workflow__bare_translation__unstructured_horizontal_has_unit_stride=unstructured_horizontal_has_unit_stride,
        otf_workflow__bare_translation__use_metrics=use_metrics,
        otf_workflow__bare_translation__disable_field_origin_on_program_arguments=use_zero_origin,
        otf_workflow__bare_translation__use_max_domain_range_on_unstructured_shift=use_max_domain_range_on_unstructured_shift,
    )


run_dace_cpu = make_dace_backend(
    gpu=False,
    cached=False,
    auto_optimize=True,
    async_sdfg_call=False,
)
run_dace_cpu_noopt = make_dace_backend(
    gpu=False,
    cached=False,
    auto_optimize=False,
    async_sdfg_call=False,
)
run_dace_cpu_cached = make_dace_backend(
    gpu=False,
    cached=True,
    auto_optimize=True,
    async_sdfg_call=False,
)

run_dace_gpu = make_dace_backend(
    gpu=True,
    cached=False,
    auto_optimize=True,
    async_sdfg_call=True,
)
run_dace_gpu_noopt = make_dace_backend(
    gpu=True,
    cached=False,
    auto_optimize=False,
    async_sdfg_call=True,
)
run_dace_gpu_cached = make_dace_backend(
    gpu=True,
    cached=True,
    auto_optimize=True,
    async_sdfg_call=True,
)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import functools
import inspect
import pathlib
import symtable
import textwrap
from collections.abc import Callable, Iterator
from dataclasses import dataclass
from typing import Any, cast


MISSING_FILENAME = "<string>"


def get_closure_vars_from_function(function: Callable) -> dict[str, Any]:
    (nonlocals, globals, builtins, _unbound) = inspect.getclosurevars(function)  # noqa: A001 [builtin-variable-shadowing]

    # nonlocals override globals, sorted for deterministic results
    return dict(sorted({**builtins, **globals, **nonlocals}.items()))


def make_source_definition_from_function(func: Callable) -> SourceDefinition:
    try:
        filename = str(pathlib.Path(inspect.getabsfile(func)).resolve())
        if not filename:
            raise ValueError(
                "Can not create field operator from a function that is not in a source file."
            )
        source_lines, line_offset = inspect.getsourcelines(func)
        source_code = textwrap.dedent(inspect.getsource(func))
        column_offset = min(
            [len(line) - len(line.lstrip()) for line in source_lines if line.lstrip()], default=0
        )
        return SourceDefinition(source_code, filename, line_offset - 1, column_offset)

    except OSError as err:
        raise ValueError(f"Can not get source code of passed function '{func}'.") from err


def make_symbol_names_from_source(source: str, filename: str = MISSING_FILENAME) -> SymbolNames:
    try:
        mod_st = symtable.symtable(source, filename, "exec")
    except SyntaxError as err:
        raise ValueError(
            f"Unexpected error when parsing provided source code: \n{source}\n"
        ) from err

    assert mod_st.get_type() == "module"
    if len(children := mod_st.get_children()) != 1:
        raise ValueError(
            f"Sources with multiple function definitions are not yet supported: \n{source}\n"
        )

    assert children[0].get_type() == "function"
    func_st: symtable.Function = cast(symtable.Function, children[0])

    param_names: set[str] = set()
    imported_names: set[str] = set()
    local_names: set[str] = set()
    for name in func_st.get_locals():
        if (s := func_st.lookup(name)).is_imported():
            imported_names.add(name)
        elif s.is_parameter():
            param_names.add(name)
        else:
            local_names.add(name)

    # symtable returns regular free (or non-local) variables in 'get_frees()' and
    # the free variables introduced with the 'nonlocal' statement in 'get_nonlocals()'
    nonlocal_names = set(func_st.get_frees()) | set(func_st.get_nonlocals())
    global_names = set(func_st.get_globals())

    return SymbolNames(
        params=param_names,
        locals=local_names,
        imported=imported_names,
        nonlocals=nonlocal_names,
        globals=global_names,
    )


@dataclass(frozen=True)
class SourceDefinition:
    """
    A GT4Py source code definition encoded as a string.

    It can be created from an actual function object using :meth:`from_function()`.
    It also supports unpacking.


    Examples
    --------
    >>> def foo(a):
    ...     return a
    >>> src_def = SourceDefinition.from_function(foo)
    >>> print(src_def)  # doctest:+ELLIPSIS
    SourceDefinition(source='def foo(a):...', filename='...', line_offset=0, column_offset=0)

    >>> source, filename, starting_line = src_def
    >>> print(source)  # doctest:+ELLIPSIS
    def foo(a):
        return a
    ...
    """

    source: str
    filename: str = MISSING_FILENAME
    line_offset: int = 0
    column_offset: int = 0

    def __iter__(self) -> Iterator:
        yield self.source
        yield self.filename
        yield self.line_offset

    from_function = staticmethod(make_source_definition_from_function)


@dataclass(frozen=True)
class SymbolNames:
    """
    Collection of symbol names used in a function classified by kind.

    It can be created directly from source code using :meth:`from_source()`.
    It also supports unpacking.
    """

    params: set[str]
    locals: set[str]  # shadowing a python builtin
    imported: set[str]
    nonlocals: set[str]
    globals: set[str]  # shadowing a python builtin

    @functools.cached_property
    def all_locals(self) -> set[str]:
        return self.params | self.locals | self.imported

    def __iter__(self) -> Iterator[set[str]]:
        yield self.params
        yield self.locals
        yield self.imported
        yield self.nonlocals
        yield self.globals

    from_source = staticmethod(make_symbol_names_from_source)
